import re
import unicodedata
from collections import Counter

from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.formatting.plain import Style as PlainStyle
from pybtex.style.labels import BaseLabelStyle
from pybtex.plugin import register_plugin

_alphanum_ = re.compile('[^A-Za-z0-9-]+', re.UNICODE)

def strip_accents(s):
    return ''.join((c for c in unicodedata.normalize('NFD', s)
                    if not unicodedata.combining(c)))

def lastname(p):
    s = strip_accents(''.join(p.prelast_names + p.last_names))
    return _alphanum_.sub('', s)

class AuthorsLabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        labels = [self.format_label(entry) for entry in sorted_entries]
        count = Counter(labels)
        counted = Counter()
        for label in labels:
            if count[label] == 1:
                yield label
            else:
                yield label + chr(ord('a') + counted[label])
                counted.update([label])

    def format_label(self, entry):
        # see alpha.bst calc.label
        org    = entry.type in ("proceedings", "manual")
        editor = entry.type in ("book", "inbook", "proceedings",
                                "inproceedings")
        author = entry.type not in ("proceedings")
        label = self.generate_label(entry, author=author,
                                    editor=editor, org=org)
        if "year" in entry.fields:
            label = "{:s} ({:s})".format(label, entry.fields["year"])
        return label

    def generate_label(self, entry, author=True, editor=True, org=True):
        result = entry.fields["key"] if "key" in entry.fields else entry.key
        if author and "author" in entry.persons:
            result = self.names2label(entry.persons["author"])
        elif editor and "editor" in entry.persons:
            result = self.names2label(entry.persons["editor"])
        elif org and "organization" in entry.fields:
            result = entry.fields["organization"]
            if result.startswith("The "):
                result = result[4:]
        return result

    def names2label(self, persons):
        numnames = len(persons)
        if numnames == 1:
            result = lastname(persons[0])
        elif numnames == 2:
            result = '{:s} and {:s}'.format(lastname(persons[0]),
                                            lastname(persons[1]))
        elif numnames == 3:
            result = '{:s}, {:s}, and {:s}'.format(lastname(persons[0]),
                                                   lastname(persons[1]),
                                                   lastname(persons[2]))
        else:
            result = '{:s} et al.'.format(lastname(persons[0]))
        return result

class AuthorListPlainStyle(PlainStyle):
    default_label_style = AuthorsLabelStyle

class AuthorListUnsrtStyle(UnsrtStyle):
    default_label_style = AuthorsLabelStyle

register_plugin('pybtex.style.formatting', 'authorlist', AuthorListPlainStyle)
register_plugin('pybtex.style.formatting', 'authorlist_unsrt', AuthorListUnsrtStyle)



#!/usr/bin/env python
""" Tests of xraydb interface  """
import time
import pytest
import numpy as np
from numpy.testing import assert_allclose

from xraydb import XrayDB

def test_xraydb_version():
    xdb = XrayDB()
    version = xdb.get_version()
    assert 'XrayDB Version: 8.1, Python' in version

    hist = xdb.get_version(with_history=True)
    assert len(hist) > 350
    hist = hist.split('\n')
    assert len(hist) > 4

def test_molar_mass():
    xdb = XrayDB()
    assert_allclose(xdb.molar_mass('Co'), 58.932, rtol=0.001)

    with pytest.raises(ValueError):
        xdb.molar_mass('Mx')

def test_density():
    xdb = XrayDB()
    assert_allclose(xdb.density('Ne'), 0.00090, rtol=0.01)
    assert_allclose(xdb.density('Ti'), 4.506,   rtol=0.01)
    assert_allclose(xdb.density('Kr'), 0.00375, rtol=0.01)
    assert_allclose(xdb.density('Mo'), 10.28,   rtol=0.01)
    assert_allclose(xdb.density('Pd'), 12.03,   rtol=0.01)
    assert_allclose(xdb.density('Au'), 19.3,    rtol=0.01)
    with pytest.raises(ValueError):
        xdb.density('Mx')

def test_atomic_number():
    xdb = XrayDB()
    assert xdb.atomic_number('Ne') == 10

    with pytest.raises(ValueError):
        xdb.atomic_number('Mx')

def test_symbol():
    xdb = XrayDB()
    assert xdb.symbol(40) == 'Zr'

def test_xray_line_strengths():
    xdb = XrayDB()
    assert len(xdb.xray_line_strengths('Hg', excitation_energy=2800)) == 2
    assert len(xdb.xray_line_strengths('Hg', excitation_energy=12000)) == 3
    assert len(xdb.xray_line_strengths('Hg', excitation_energy=12500)) == 9
    assert len(xdb.xray_line_strengths('Hg', excitation_energy=14300)) == 13
    assert len(xdb.xray_line_strengths('Hg', excitation_energy=16000)) == 17

def test_ionization_potentials():
    xdb = XrayDB()
    assert xdb.ionization_potential('air') == 33.8
    assert xdb.ionization_potential('helium') == 41.3
    assert xdb.ionization_potential('He') == 41.3
    assert xdb.ionization_potential('Si') == 3.68

    with pytest.raises(ValueError):
        xdb.ionization_potential('p10')



from xraydb import XrayDB
db = XrayDB()

def extract(dat, key):
    val = dat.get(key, [' '])[0]
    if isinstance(val, float) and val > 1:
        val = int(round(val))
    return str(val)

elnames = ('', 'hydrogen' , 'helium', 'lithium' , 'beryllium', 'boron' ,
           'carbon' ,'nitrogen' , 'oxygen', 'fluorine' , 'neon' ,'sodium',
           'magnesium', 'aluminum' , 'silicon', 'phosphorus' , 'sulfur',
           'chlorine' , 'argon', 'potassium' , 'calcium', 'scandium' ,
           'titanium','vanadium' , 'chromium','manganese' , 'iron',
           'cobalt' , 'nickel','copper' , 'zinc','gallium' ,
           'germanium','arsenic' , 'selenium','bromine' , 'krypton',
           'rubidium' , 'strontium','yttrium' , 'zirconium','niobium' ,
           'molybdenum','technetium' , 'ruthenium','rhodium' ,
           'palladium','silver' , 'cadmium','indium' , 'tin', 'antimony' ,
           'tellurium','iodine' , 'xenon','cesium' , 'barium','lanthanum' ,
           'cerium','praseodymium', 'neodymium', 'promethium' ,
           'samarium','europium' , 'gadolinium', 'terbium' ,
           'dysprosium','holmium' , 'erbium','thulium' ,
           'ytterbium','lutetium' , 'hafnium','tantalum' , 'tungsten',
           'rhenium' , 'osmium','iridium' , 'platinum','gold' ,
           'mercury','thallium' , 'lead','bismuth' , 'polonium', 'astatine',
           'radon','francium' , 'radium','actinium' ,
           'thorium','protactinium', 'uranium','neptunium' , 'plutonium',
           'americium', 'curium', 'berkelium', 'californium')

chemstates = {'':'',
              'hydrogen':'+1',
              'helium':'',
              'lithium' :'+1',
              'beryllium':'+2',
              'boron' :'+3',
              'carbon' :'-4, -3, \ldots, +2, +3, +4',
              'nitrogen' :'-3, +3, +5',
              'oxygen':'-2',
              'fluorine' :'-1',
              'neon' :'',
              'sodium':'+1',
              'magnesium':'+2',
              'aluminum' :'+3',
              'silicon':'-4, +4',
              'phosphorus' :'-3, +3, +5',
              'sulfur':'-2, +2, +4, +6',
              'chlorine' :'-1, +1, +3, +5, +7',
              'argon':'',
              'potassium' :'+1',
              'calcium':'+2',
              'scandium' :'+3',
              'titanium':'+3, +4',
              'vanadium' :'+2, +3, +4, +5',
              'chromium':'+2, +3, +6',
              'manganese' :'+2, +3, +4, +7',
              'iron':'+2, +3',
              'cobalt' :'+2, +3',
              'nickel':'+2',
              'copper' :'+1, +2',
              'zinc':'+2',
              'gallium' :'+3',
              'germanium':'-4, +2, +4',
              'arsenic' :'-3, +3, +5',
              'selenium':'-2, +2, +4, +6',
              'bromine' :'-1, +1, +3, +5',
              'krypton':'',
              'rubidium' :'+1',
              'strontium':'+2',
              'yttrium' :'+3',
              'zirconium':'+4',
              'niobium' :'+4, +5',
              'molybdenum':'+3, +4, +6',
              'technetium' :'+4, +7',
              'ruthenium':'+3, +4, +6',
              'rhodium' :'+2, +3, +4',
              'palladium':'+2, +4',
              'silver' :'+1',
              'cadmium':'+2',
              'indium' :'+3',
              'tin':'-4, +2, +4',
              'antimony' :'-3, +3, +5',
              'tellurium':'-2, +2, +4, +6',
              'iodine' :'-1, +1, +3, +5, +7',
              'xenon':'',
              'cesium' :'+1',
              'barium':'+2',
              'lanthanum' :'+3',
              'cerium':'+3, +4',
              'praseodymium':'+3, +4',
              'neodymium':'+3',
              'promethium' :'+3',
              'samarium':'+3',
              'europium' :'+2, +3',
              'gadolinium':'+3',
              'terbium' :'+3, +4',
              'dysprosium':'+3',
              'holmium' :'+3',
              'erbium':'+3',
              'thulium' :'+3',
              'ytterbium':'+3',
              'lutetium' :'+3',
              'hafnium':'+4',
              'tantalum' :'+5',
              'tungsten':'+4, +6',
              'rhenium' :'+4',
              'osmium':'+4',
              'iridium' :'+3, +4',
              'platinum':'+2, +4',
              'gold' :'+1, +3',
              'mercury':'+1, +2',
              'thallium' :'+1, +3',
              'lead':'+2, +4',
              'bismuth' :'+3, +5',
              'polonium':'-2, +2, +4',
              'astatine':'-1, +1',
              'radon':'',
              'francium' :'+1',
              'radium':'+2',
              'actinium' :'+3',
              'thorium':'+4',
              'protactinium':'+5',
              'uranium':'+4, +6',
              'neptunium' :'+3, +4, +5',
              'plutonium':'+3, +4, +5',
              'americium':'+3, +4, +5',
              'curium':'+3',
              'berkelium':'+3, +4',
              'californium':'+3'}


table=r"""\newcommand{\%(texsym)s}{{%%
\begin{minipage}{67mm}

\vspace{1mm}

{\Huge{\hspace{1mm} {\textbf{%(sym)s}} \hfill \hfil{\textbf{%(iz)s}} \hspace{1mm}}} %%

\vspace{4mm}

{\Huge{\hfill {\Name{%(name)s}} \hfill}}

\vspace{5mm}

{\Large{
\begin{tabular*}{67mm}%%
{@{\hspace{5pt}}{r}@{\extracolsep{\fill}}r@{\extracolsep{\fill}}r}%%
{\BRed{%(k)s}}  & {\textbf{%(ka1)s}} &  {\textbf{%(kb1)s}} \\%%
{\BBlue{%(l1)s}} & %(lb3)s & %(lb4)s \\%%
{\BBlue{%(l2)s}} & %(lb1)s & %(lg1)s \\%%
{\BRed{%(l3)s}} & {\textbf{%(la1)s}} & {\textbf{%(lb2)s}} \\%%
{\BBlue{%(m5)s}} & %(ma)s & %(mb)s \\%%
\noalign{\medskip} %%  \multicolumn{3}{@{\hspace{1pt}}c}{ }\\%%
{\textbf{%(mass)s}} & \multicolumn{2}{r}{\large{%(states)s}}\\%%
\end{tabular*}
}}
\end{minipage}}}
"""

for iz in range(1, 99):
    sym = db.symbol(iz)

    dat = {'iz': iz, 'sym': sym, 'name': elnames[iz],
           'texsym': "Elem%s" % sym}

    edges = db.xray_edges(iz)
    lines = db.xray_lines(iz)
    dat['mass']  = str(db.molar_mass(iz))
    dat['states']  = chemstates.get(elnames[iz], '')


    dat['k']   = extract(edges, 'K')
    dat['ka1'] = extract(lines, 'Ka1')
    dat['kb1'] = extract(lines, 'Kb1')

    dat['l1']  = extract(edges, 'L1')
    dat['lb3'] = extract(lines, 'Lb3')
    dat['lb4'] = extract(lines, 'Lb4')

    dat['l2']  = extract(edges, 'L2')
    dat['lb1'] = extract(lines, 'Lb1')
    dat['lg1'] = extract(lines, 'Lg1')

    dat['l3']  = extract(edges, 'L3')
    dat['la1'] = extract(lines, 'La1')
    dat['lb2'] = extract(lines, 'Lb2,15')

    dat['m5'] = extract(edges, 'M5')
    dat['ma'] = extract(lines, 'Ma')
    dat['mb'] = extract(lines, 'Mb')

    # print dat.keys()
    print('%% ', dat['name'])
    print(table % dat)

highz  = ((99, 'Es', 'einsteinium'),
         (100, 'Fm', 'fermium'),
         (101, 'Md', 'mendelevium'),
         (102, 'No', 'nobelium'),
         (103, 'Lr', 'lawrencium'))

for iz, sym, name in highz:
    dat = {'iz': iz, 'sym': sym, 'name': name,
           'texsym': "Elem%s" % sym,
           'k': '', 'ka1': '', 'kb1': '', 'l1' : '', 'lb3': '', 'lb4': '',
           'l2' : '', 'lb1': '', 'lg1': '', 'l3' : '', 'la1': '', 'lb2':
           '', 'm5': '', 'ma': '', 'mb': '', 'mass':'', 'states':''}

    print( '%% ', dat['name'])
    print(table % dat)

print('%% Key')
print(table % {'iz': 'Z', 'sym': 'Symbol', 'name': 'name',
               'mass': 'Mass',  'states': 'oxidation states',
               'texsym': "ElemKey",
               'k':   r'$\mathbf{K}$ edge',
               'ka1': r'$\mathbf{K_{\alpha_1}}$',
               'kb1': r'$\mathbf{K_{\beta_1}}$',
               'l1':  r'$\mathrm{L_{\rm 1}}$ edge',
               'lb3': r'$\mathrm{L_{\beta_3}}$',
               'lb4': r'$\mathrm{L_{\beta_4}}$',
               'l2':  r'$\mathrm{L_{\rm 2}}$ edge',
               'lb1': r'$\mathrm{L_{\beta_1}}$',
               'lg1': r'$\mathrm{L_{\gamma_1}}$',
               'l3':  r'$\mathbf{L_{\rm 3}}$ edge',
               'la1': r'$\mathbf{L_{\alpha_1}}$',
               'lb2': r'$\mathbf{L_{\beta_2}}$',
               'm5':  r'$\mathrm{M_{\rm 5}}$ edge',
               'ma':  r'$\mathrm{M_{\alpha}}$',
               'mb':  r'$\mathrm{M_{\beta}}$'})



import numpy as np
from xraydb import darwin_width
import matplotlib.pyplot as plt

dw_si111 = darwin_width(10000, 'Si', (1, 1, 1))
dw_si333 = darwin_width(30000, 'Si', (3, 3, 3))

fmt_string = "Darwin Width for {:s} at {:.0f} keV: {:5.2f} microrad, {:5.2f} eV"
print(fmt_string.format('Si(111)', 10,
                        dw_si111.theta_width*1e6,
                        dw_si111.energy_width))


print(fmt_string.format('Si(333)', 30,
                        dw_si333.theta_width*1e6,
                        dw_si333.energy_width))

dtheta  = dw_si111.dtheta*1e6
denergy = dw_si111.denergy[::-1]

#  slightly advanced matplotlib hackery:
fig, ax = plt.subplots(constrained_layout=True)


ax.plot(dtheta, dw_si111.intensity, label='$I$, Si(111)', linewidth=2)
ax.plot(dtheta, dw_si111.intensity**2, label='$I^2$, Si(111)', linewidth=2)
ax.plot(dw_si333.dtheta*1e6, dw_si333.intensity**2, label='$I^2$ Si(333) 30 keV', linewidth=2)


ax.set_title('X-ray diffraction intensity at 10keV')
ax.set_xlabel('Angle - $\\theta_B$ ($ \mu \mathrm{rad}$)')
ax.set_ylabel('Reflectivity')
ax.legend()

plt.show()



#!/usr/bin/env python
""" Tests of Larch Scripts  """
import os
import time
import shutil
import pytest
import numpy as np
from numpy.testing import assert_allclose

from xraydb import (chemparse, validate_formula, material_mu,
                    material_mu_components, find_material, get_materials,
                    get_material, add_material)

from xraydb.materials import get_user_materialsfile

def test_chemparse():
    examples = {'H2O': {'H':2, 'O':1},
                'Mn(SO4)2(H2O)7':  {'H': 14.0, 'S': 2.0, 'Mn': 1, 'O': 15.0},
                'Mg(SO4)2': {'Mg': 1, 'S': 2, 'O':8},
                'Mg0.5Fe0.5' : {'Mg': 0.5, 'Fe': 0.5},
                'Mg.3Fe.7' : {'Mg': 0.3, 'Fe': 0.7},
                'CO': {'C': 1, 'O': 1} }
    for formula, cert in examples.items():
        ret = chemparse(formula)
        for elem, quant in cert.items():
            v = ret.pop(elem)
            assert_allclose(v, quant, rtol=1.e3)
        assert len(ret)==0

def test_validate_formula():
    examples = {'H2O': True,
                'Mn(SO4)2(H2O)7':  True,
                'Mn(SO42(H2O)7': False,
                'Z': False}
    for formula, cert in examples.items():
        ret = validate_formula(formula)
        assert (ret == cert)

def test_get_materials():
    examples = {'water': True, 'lead': True, 'acetone': True,
                'kapton': True, 'sapphire': True,
                'cheese': False, 'spice': False, 'bicycle': False}

    known_materials = get_materials()
    for name, found in examples.items():
        if found:
            assert (name in known_materials)
        else:
            assert (name not in known_materials)

def test_material_mu1():
    en = np.linspace(8500, 9500, 21)
    known_mu = np.array([236.2, 232.4, 228.7, 225.1, 221.5, 218.1, 214.7,
                         211.4, 208.1, 204.9, 1403.6, 1385.6, 1367.9, 1350.6,
                         1333.5, 1316.7, 1300.3, 1284.0, 1268.1, 1252.4,
                         1237.0])

    mu = material_mu('CuO', en, density=6.3)
    assert_allclose(mu, known_mu, rtol=0.005)

def test_material_mu2():
    en = np.linspace(5000, 10000, 21)
    known_mu = np.array([0.04934, 0.04267, 0.03715, 0.03254, 0.02866,
                         0.02538, 0.02257, 0.02016, 0.01809, 0.01629,
                         0.01472, 0.01334, 0.01214, 0.01108, 0.01014,
                         0.00930, 0.00856, 0.00789, 0.00729, 0.00675,
                         0.00626])

    mu = material_mu('air', en)
    assert_allclose(mu, known_mu, rtol=0.05)

    air_formula, air_density = get_material('air')

    air_comps = chemparse(air_formula)

    assert air_comps['Ar'] < 0.013
    assert air_comps['Ar'] > 0.007

    mu = material_mu('air', en, density=2.0*air_density)
    assert_allclose(mu, 2.0*known_mu, rtol=0.05)

def test_material_mu3():
    en = np.linspace(5000, 10000, 21)

    known_mu = np.array([42.592, 36.801, 32.006, 28.005, 24.641, 21.794,
                         19.367, 17.288, 15.496, 13.943, 12.592, 11.411,
                         10.373, 9.458, 8.649, 7.931, 7.291, 6.719, 6.206,
                         5.745, 5.330])

    mu = material_mu('H2O', en)
    assert_allclose(mu, known_mu, rtol=0.05)

def test_material_mu4():
    en = np.linspace(5000, 10000, 21)

    known_mu = np.array([42.592, 36.801, 32.006, 28.005, 24.641, 21.794,
                         19.367, 17.288, 15.496, 13.943, 12.592, 11.411,
                         10.373, 9.458, 8.649, 7.931, 7.291, 6.719, 6.206,
                         5.745, 5.330])

    with pytest.raises(Warning):
        out = material_mu('H2SO4', en)


def test_material_mu_components1():
    mu = material_mu('quartz', 10000)
    assert_allclose(mu, 50.368, rtol=0.001)

    comps = material_mu_components('quartz', 10000)

    known_comps =  {'mass': 60.08, 'density': 2.65, 'elements': ['Si', 'O'],
                    'Si': (1, 28.1, 33.879), 'O': (2.0, 16.0, 5.953)}


    assert 'Si'in comps['elements']
    assert 'O'in comps['elements']

    for attr in ('mass', 'density'):
        assert_allclose(comps[attr], known_comps[attr], rtol=0.01)

    for attr in ('Si', 'O'):
        assert_allclose(comps[attr][0], known_comps[attr][0], rtol=0.01)
        assert_allclose(comps[attr][1], known_comps[attr][1], rtol=0.01)
        assert_allclose(comps[attr][2], known_comps[attr][2], rtol=0.01)


def test_material_mu_components2():
    mu = material_mu('TiO2', 10000, density=4.23)
    assert_allclose(mu, 290.7, rtol=0.001)

    mu = material_mu('TiO2', 10000)
    assert_allclose(mu, 290.7, rtol=0.001)

    mu = material_mu('TiO2', 10000, density=4.5)
    assert_allclose(mu, 309.26, rtol=0.001)

    comps = material_mu_components('TiO2', 10000, density=4.23)


    known_comps =  {'mass': 79.88, 'density': 4.23, 'elements': ['Ti', 'O'],
                    'Ti': (1, 47.88, 110.676), 'O': (2.0, 15.9994, 5.953)}

    assert 'Ti'in comps['elements']
    assert 'O'in comps['elements']

    for attr in ('mass', 'density'):
        assert_allclose(comps[attr], known_comps[attr], rtol=0.01)

    for attr in ('Ti', 'O'):
        assert_allclose(comps[attr][0], known_comps[attr][0], rtol=0.01)
        assert_allclose(comps[attr][1], known_comps[attr][1], rtol=0.01)
        assert_allclose(comps[attr][2], known_comps[attr][2], rtol=0.01)

    with pytest.raises(Warning):
        c = material_mu_components('TiO2', 10000)


def test_material_find():
    mat_  = {'kapton': ('C22H10N2O5', 1.43, 'polymer'),
             'lead': ('Pb', 11.34, 'metal'),
             'aluminum': ('Al', 2.72, 'metal'),
             'water': ('H2O', 1.0, 'solvent')}
    for mname, mdat in mat_.items():
        mat = find_material(mname)
        f1 = chemparse(mat.formula)
        f2 = chemparse(mdat[0])
        for k, v in f2.items():
            assert v == f1[k]
        assert_allclose(mat.density, mdat[1], rtol=0.1)
        assert(mdat[2] in mat.categories)

    for formula in ('WSO3', 'CdAs140CO3', 'KAs'):
        out = find_material(formula)
        assert out == None


def test_material_get():
    mat_  = {'kapton': ('C22H10N2O5', 1.43),
             'lead': ('Pb', 11.34),
             'aluminum': ('Al', 2.72),
             'water': ('H2O', 1.0)}

    for mname in mat_.keys():
        formula, density = get_material(mname)
        f1 = chemparse(formula)
        f2 = chemparse(mat_[mname][0])
        for k, v in f2.items():
            assert v == f1[k]
        assert_allclose(density, mat_[mname][1], rtol=0.1)

    for formula, density in mat_.values():
        _f, _d = get_material(formula)
        assert _f == formula
        assert_allclose(density, _d, rtol=0.1)

    for formula in ('WSO3', 'CdAs140CO3', 'KAs'):
        out = get_material(formula)
        assert out == None


def test_material_add():
    matfile = get_user_materialsfile()
    savefile = matfile + '_Save'
    if os.path.exists(matfile):
        had_matfile = True
        shutil.move(matfile, savefile)
        time.sleep(2.0)

    add_material('caffeine', 'C8H10N4O2', density=1.23)
    assert get_material('caffeine') is not None

    time.sleep(2.0)

    add_material('rutile', 'TiO2', density=4.23)

    with open(matfile, 'r') as fh:
        text = fh.read()

    assert 'caffeine' in text
    assert 'rutile' in text

    os.unlink(matfile)
    if os.path.exists(savefile):
        shutil.move(savefile, matfile)



#!/usr/bin/env python
"""
SQLAlchemy wrapping of x-ray database for data from
     Elam et al, Chantler et al, Waasmaier and Kirfel

Main Class for full Database:  xrayDB
"""

import os
import json
from collections import namedtuple
import numpy as np
from scipy.interpolate import UnivariateSpline

from sqlalchemy import MetaData, create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import SingletonThreadPool

from .utils import elam_spline, as_ndarray

XrayEdge = namedtuple('XrayEdge', ('energy', 'fyield', 'jump_ratio'))
XrayLine = namedtuple('XrayLine', ('energy', 'intensity', 'initial_level',
                                   'final_level'))
ElementData = namedtuple('ElementData', ('Z', 'symbol', 'mass', 'density'))

__version__ = '1.4'

def make_engine(dbname):
    "create engine for sqlite connection"
    return create_engine('sqlite:///%s' % (dbname),
                         poolclass=SingletonThreadPool,
                         connect_args={'check_same_thread': False})

def isxrayDB(dbname):
    """whether a file is a valid XrayDB database

    Args:
        dbname (string): name of XrayDB file

    Returns:
        bool: is file a valid XrayDB

    Notes:
      1. must be a sqlite db file, with tables named 'elements',
        'photoabsorption', 'scattering', 'xray_levels', 'Coster_Kronig',
        'Chantler', 'Waasmaier', and 'KeskiRahkonen_Krause'
    """
    _tables = ('Chantler', 'Waasmaier', 'Coster_Kronig',
               'KeskiRahkonen_Krause', 'xray_levels',
               'elements', 'photoabsorption', 'scattering')
    result = False
    try:
        engine = make_engine(dbname)
        meta = MetaData(engine)
        meta.reflect()
        result = all([t in meta.tables for t in _tables])
    except:
        pass
    return result


class XrayDB():
    """
    Database of Atomic and X-ray Data

    This XrayDB object gives methods to access the Atomic and
    X-ray data in th SQLite3 database xraydb.sqlite.

    Much of the data in this database comes from the compilation
    of Elam, Ravel, and Sieber, with additional data from Chantler,
    and other sources. See the documention and bibliography for
    a complete listing.
    """

    def __init__(self, dbname='xraydb.sqlite', read_only=True):
        "connect to an existing database"
        if not os.path.exists(dbname):
            parent, _ = os.path.split(__file__)
            dbname = os.path.join(parent, dbname)
            if not os.path.exists(dbname):
                raise IOError("Database '%s' not found!" % dbname)

        if not isxrayDB(dbname):
            raise ValueError("'%s' is not a valid X-ray Database file!" % dbname)

        self.dbname = dbname
        self.engine = make_engine(dbname)
        self.conn = self.engine.connect()
        kwargs = {}
        if read_only:
            kwargs = {'autoflush': True, 'autocommit': False}

            def readonly_flush(*args, **kwargs):
                return

            self.session = sessionmaker(bind=self.engine, **kwargs)()
            self.session.flush = readonly_flush
        else:
            self.session = sessionmaker(bind=self.engine, **kwargs)()

        self.metadata = MetaData(self.engine)
        self.metadata.reflect()
        self.tables = self.metadata.tables
        elems = self.tables['elements'].select().execute()
        self.atomic_symbols = [e.element for e in elems.fetchall()]

    def close(self):
        "close session"
        self.session.flush()
        self.session.close()

    def query(self, *args, **kws):
        "generic query"
        return self.session.query(*args, **kws)

    def get_version(self, long=False, with_history=False):
        """
        return sqlite3 database and python library version numbers

        Parameters:
            long (bool): show timestamp and notes of latest version [False]
            with_history (bool): show complete version history [False]

        Returns:
            string: version information
        """
        out = []
        rows = self.tables['Version'].select().execute().fetchall()
        if not with_history:
            rows = rows[-1:]
        if long or with_history:
            for row in rows:
                out.append("XrayDB Version: %s [%s] '%s'" % (row.tag,
                                                             row.date,
                                                             row.notes))
            out.append("Python Version: %s" % __version__)
            out = "\n".join(out)
        else:
            out = "XrayDB Version: %s, Python Version: %s" % (rows[0].tag,
                                                              __version__)
        return out

    def f0_ions(self, element=None):
        """
        return list of ion names supported for the .f0() function.


        Parameters:
            element (string, int, pr None):  atomic number, symbol, or ionic symbol
                    of scattering element.

        Returns:
            list:  if element is None, all 211 ions are returned.
                   if element is not None, the ions for that element are returned

        Example:
            >>> xdb = XrayDB()
            >>> xdb.f0_ions('Fe')
            ['Fe', 'Fe2+', 'Fe3+']

        Notes:
            Z values from 1 to 98 (and symbols 'H' to 'Cf') are supported.

        References:
            Waasmaier and Kirfel
        """
        wtab = self.tables['Waasmaier']
        rows = self.query(wtab)
        if element is not None:
            elem = self.symbol(element)
            rows = rows.filter(wtab.c.element == elem)
        return [str(r.ion) for r in rows.all()]

    def f0(self, ion, q):
        """
        return f0(q) -- elastic X-ray scattering factor from Waasmaier and Kirfel

        Parameters:
            ion (string, int, or None):  atomic number, symbol or ionic symbol
                  of scattering element.
            q (float, list, ndarray): value(s) of q for scattering factors

        Returns:
            ndarray: elastic scattering factors


        Example:
            >>> xdb = XrayDB()
            >>> xdb.f0('Fe', range(10))
            array([ 25.994603  ,   6.55945765,   3.21048827,   1.65112769,
                     1.21133507,   1.0035555 ,   0.81012185,   0.61900285,
                     0.43883403,   0.27673021])

        Notes:
            q = sin(theta) / lambda, where theta = incident angle,
            and lambda = X-ray wavelength

            Z values from 1 to 98 (and symbols 'H' to 'Cf') are supported.
            The list of ionic symbols can be read with the function .f0_ions()

        References:
            Waasmaier and Kirfel
        """
        wtab = self.tables['Waasmaier']
        if isinstance(ion, int):
            row = self.query(wtab).filter(wtab.c.atomic_number == ion).all()[0]
        elif ion not in self.f0_ions():
            raise ValueError('No ion {:s} from Waasmaier table'.format(repr(ion)))
        else:
            row = self.query(wtab).filter(wtab.c.ion == ion.title()).all()[0]
        q = as_ndarray(q)
        f0 = row.offset
        for s, e in zip(json.loads(row.scale), json.loads(row.exponents)):
            f0 += s * np.exp(-e*q*q)
        return f0

    def _from_chantler(self, element, energy, column='f1', smoothing=0):
        """
        return energy-dependent data from Chantler table

        Parameters:
            element (string or int): atomic number or symbol.
            eneregy (float or ndarray):
        columns: f1, f2, mu_photo, mu_incoh, mu_total

        Notes:
           this function is meant for internal use.
        """
        ctab = self.tables['Chantler']
        elem = self.symbol(element)
        row = self.query(ctab).filter(ctab.c.element == elem).one()

        energy = as_ndarray(energy)
        emin, emax = min(energy), max(energy)

        te = np.array(json.loads(row.energy))
        nemin = max(0, -3 + max(np.where(te <= emin)[0]))
        nemax = min(len(te), 3 + max(np.where(te <= emax)[0]))

        te = te[nemin:nemax+1]
        if column == 'mu':
            column = 'mu_total'
        ty = np.array(json.loads(getattr(row, column)))[nemin:nemax+1]
        ty[np.where(abs(ty) < 1.e-99)] =  1.e-99
        if column == 'f1':
            out = UnivariateSpline(te, ty, s=smoothing)(energy)
        else:
            out = np.exp(np.interp(np.log(energy),
                                   np.log(te),
                                   np.log(ty)))
        if isinstance(out, np.ndarray) and len(out) == 1:
            out = out[0]
        return out

    def chantler_energies(self, element, emin=0, emax=1.e9):
        """
        return array of energies (in eV) at which data is
        tabulated in the Chantler tables for a particular element.

        Parameters:
            element (string or int): atomic number or symbol
            emin (float): minimum energy (in eV) [0]
            emax (float): maximum energy (in eV) [1.e9]

        Returns:
            ndarray: energies

        References:
            Chantler

        Notes:
            returns 2 energies below emin and above emax to better
            enable interpolation
        """
        ctab = self.tables['Chantler']
        elem = self.symbol(element)
        row = self.query(ctab).filter(ctab.c.element == elem).one()
        te = np.array(json.loads(row.energy))

        if emin <= min(te):
            nemin = 0
        else:
            nemin = max(0, -1 + max(np.where(te <= emin)[0]))
        if emax > max(te):
            nemax = len(te)
        else:
            nemax = min(len(te), 2 + max(np.where(te <= emax)[0]))
        return te[nemin:nemax+1]

    def f1_chantler(self, element, energy, **kws):
        """
        returns f1 -- real part of anomalous X-ray scattering factor
        for selected input energy (or energies) in eV.

        Parameters:
            element (string or int): atomic number or symbol
            energy (float or ndarray): energies (in eV).

        Returns:
            ndarray: real part of anomalous scattering factor

        References:
            Chantler
        """
        return self._from_chantler(element, energy, column='f1', **kws)

    def f2_chantler(self, element, energy, **kws):
        """
        returns f2 -- imaginary part of anomalous X-ray scattering factor
        for selected input energy (or energies) in eV.

        Parameters:
            element (string or int): atomic number or symbol
            energy (float or ndarray): energies (in eV).

        Returns:
            ndarray: imaginary part of anomalous scattering factor

        References:
            Chantler
        """
        return self._from_chantler(element, energy, column='f2', **kws)

    def mu_chantler(self, element, energy, incoh=False, photo=False):
        """
        returns X-ray mass attenuation coefficient, mu/rho in cm^2/gr
        for selected input energy (or energies) in eV.
        default is to return total attenuation coefficient.

        Parameters:
            element (string or int): atomic number or symbol
            energy (float or ndarray): energies (in eV).
            photo (bool): return only the photo-electric contribution [False]
            incoh (bool): return only the incoherent contribution [False]

        Returns:
            ndarray: mass attenuation coefficient in cm^2/gr

        References:
            Chantler
        """
        col = 'mu_total'
        if photo:
            col = 'mu_photo'
        elif incoh:
            col = 'mu_incoh'
        return self._from_chantler(element, energy, column=col)

    def _elem_data(self, element):
        "return data from elements table: internal use"
        etab = self.tables['elements']
        row = self.query(etab)
        if isinstance(element, int):
            row = row.filter(etab.c.atomic_number == element).one()
        else:
            elem = element.title()
            if not elem in self.atomic_symbols:
                raise ValueError("unknown element '%s'" % repr(elem))
            row = row.filter(etab.c.element == elem).one()
        return ElementData(int(row.atomic_number),
                           row.element.title(),
                           row.molar_mass, row.density)

    def atomic_number(self, element):
        """
        return element's atomic number

        Parameters:
            element (string or int): atomic number or symbol

        Returns:
            integer: atomic number
        """
        return self._elem_data(element).Z

    def symbol(self, element):
        """
        return element symbol

        Parameters:
            element (string or int): atomic number or symbol

        Returns:
            string: element symbol
        """
        return self._elem_data(element).symbol

    def molar_mass(self, element):
        """
        return molar mass of element

        Parameters:
            element (string or int): atomic number or symbol

        Returns:
            float: molar mass of element in amu
        """
        return self._elem_data(element).mass

    def density(self, element):
        """
        return density of pure element

        Parameters:
            element (string or int): atomic number or symbol

        Returns:
            float: density of element in gr/cm^3
        """
        return self._elem_data(element).density

    def xray_edges(self, element):
        """
        returns dictionary of X-ray absorption edge energy (in eV),
        fluorescence yield, and jump ratio for an element.

        Parameters:
            element (string or int): atomic number or symbol

        Returns:
            dictionary:  keys of edge (iupac symbol), and values of
                         XrayEdge namedtuple of (energy, fyield, edge_jump)

        References:
           Elam, Ravel, and Sieber.
        """
        elem = self.symbol(element)
        ltab = self.tables['xray_levels']
        out = {}
        for r in self.query(ltab).filter(ltab.c.element == elem).all():
            out[str(r.iupac_symbol)] = XrayEdge(r.absorption_edge,
                                                r.fluorescence_yield,
                                                r.jump_ratio)
        return out

    def xray_edge(self, element, edge):
        """
        returns XrayEdge for an element and edge

        Parameters:
            element (string or int): atomic number or symbol
            edge (string):  X-ray edge

        Returns:
            XrayEdge:  namedtuple of (energy, fyield, edge_jump)

        Example:
            >>> xdb = XrayDB()
            >>> xdb.xray_edge('Co', 'K')
            XrayEdge(edge=7709.0, fyield=0.381903, jump_ratio=7.796)

        References:
           Elam, Ravel, and Sieber.
        """
        return self.xray_edges(element).get(edge.title(), None)

    def xray_lines(self, element, initial_level=None, excitation_energy=None):
        """
        returns dictionary of X-ray emission lines of an element, with

        Parameters:
            initial_level (string or list/tuple of string):  initial level(s) to
                 limit output.
            excitation_energy (float): energy of excitation, limit output those
                 excited by X-rays of this energy (in eV).

        Returns:
            dictionary: keys of lines (Siegbahn symbol), values of Xray Lines

        Notes:
            if both excitation_energy and initial_level are given, excitation_level
            will limit output

        Example:
            >>> xdb = XrayDB()
            >>> for key, val in xdb.xray_lines('Ga', 'K').items():
            >>>      print(key, val)
            'Ka3', XrayLine(energy=9068.0, intensity=0.000326203,
                            initial_level=u'K', final_level=u'L1')
            'Ka2', XrayLine(energy=9223.8, intensity=0.294438,
                            initial_level=u'K', final_level=u'L2')
            'Ka1', XrayLine(energy=9250.6, intensity=0.57501,
                            initial_level=u'K', final_level=u'L3')
            'Kb3', XrayLine(energy=10263.5, intensity=0.0441511,
                            initial_level=u'K', final_level=u'M2')
            'Kb1', XrayLine(energy=10267.0, intensity=0.0852337,
                            initial_level=u'K', final_level=u'M3')
            'Kb5', XrayLine(energy=10348.3, intensity=0.000841354,
                            initial_level=u'K', final_level=u'M4,5')

        References:
           Elam, Ravel, and Sieber.
        """
        elem = self.symbol(element)
        ttab = self.tables['xray_transitions']
        row = self.query(ttab).filter(ttab.c.element == elem)
        if excitation_energy is not None:
            initial_level = []
            for ilevel, dat in self.xray_edges(elem).items():
                if dat[0] < excitation_energy:
                    initial_level.append(ilevel.title())

        if initial_level is not None:
            if isinstance(initial_level, (list, tuple)):
                row = row.filter(ttab.c.initial_level.in_(initial_level))
            else:
                row = row.filter(ttab.c.initial_level == initial_level.title())
        out = {}
        for r in row.all():
            out[str(r.siegbahn_symbol)] = XrayLine(r.emission_energy, r.intensity,
                                                   r.initial_level, r.final_level)
        return out

    def xray_line_strengths(self, element, excitation_energy=None):
        """
        return the absolute line strength in cm^2/gr for all available lines

        Parameters:
            element (string or int): Atomic symbol or number for element
            excitation_energy (float): incident energy, in eV

        Returns:
            dictionary: elemental line with fluorescence cross section in cm2/gr.

        References:
           Elam, Ravel, and Sieber.
        """
        out = {}
        lines = self.xray_lines(element, excitation_energy=excitation_energy)
        for label, eline in lines.items():
            edge = self.xray_edge(element, eline.initial_level)
            if edge is None and ',' in eline.initial_level:
                ilevel, _ = eline.initial_level.split(',')
                edge = self.xray_edge(element, ilevel)
            if edge is not None:
                mu = self.mu_elam(element, [edge.energy*(0.999),
                                            edge.energy*(1.001)], kind='photo')
                out[label] = (mu[1]-mu[0]) * eline.intensity * edge.fyield
        return out

    def ck_probability(self, element, initial, final, total=True):
        """
        return Coster-Kronig transition probability for an element and
        initial/final levels

        Parameters:
            element (string or int): Atomic symbol or number for element
            initial (string):  initial level
            final (string):  final level
            total (bool): whether to return total or partial probability

        Returns:
            float: transition probability

        Example:
            >>> xdb = XrayDB()
            >>> xdb.ck_probability('Cu', 'L1', 'L3', total=True)
            0.681

        References:
           Elam, Ravel, and Sieber.
        """
        elem = self.symbol(element)
        ctab = self.tables['Coster_Kronig']

        row = self.query(ctab).filter(ctab.c.element == elem)
        row = row.filter(ctab.c.initial_level == initial.title())
        row = row.filter(ctab.c.final_level == final.title()).all()
        out = 0.0
        if len(row) > 0:
            row = row[0]
            out = row.transition_probability
            if total:
                out = row.total_transition_probability
        return out

    def corehole_width(self, element, edge=None, use_keski=False):
        """
        returns core hole width for an element and edge

        Parameters:
            element (string, integer): atomic number or symbol for element
            edge (string or None): edge for hole, return all if None
            use_keski (bool) : force use of KeskiRahkonen and Krause table for all data.

        Returns:
            float: corehole width in eV.

        Notes:
            Uses Krause and Oliver where data is available (K, L lines Z > 10)
            Uses Keski-Rahkonen and Krause otherwise

        References:
            Krause and Oliver, 1979
            Keski-Rahkonen and Krause, 1974

        """
        version_qy = self.tables['Version'].select().order_by('date')
        version_id = version_qy.execute().fetchall()[-1].id

        ctab = self.tables['corelevel_widths']
        if version_id < 4 or use_keski:
            ctab = self.tables['KeskiRahkonen_Krause']

        rows = self.query(ctab).filter(ctab.c.element == self.symbol(element))
        if edge is not None:
            rows = rows.filter(ctab.c.edge == edge.title())
        result = rows.all()
        if len(result) == 1:
            result = result[0].width
        else:
            result = [(r.edge, r.width) for r in result]
        return result


    def cross_section_elam(self, element, energies, kind='photo'):
        """
        returns Elam Cross Section values for an element and energies

        Parameters:
            element (string or int):  atomic number or symbol for element
            energies (float or ndarray): energies (in eV) to calculate cross-sections
            kind (string):  one of 'photo', 'coh', and 'incoh' for photo-absorption,
                  coherent scattering, and incoherent scattering cross sections,
                  respectively. Default is 'photo'.

        Returns:
            ndarray of scattering data

        References:
            Elam, Ravel, and Sieber.
        """
        elem = self.symbol(element)
        kind = kind.lower()
        if kind not in ('coh', 'incoh', 'photo'):
            raise ValueError('unknown cross section kind=%s' % kind)

        stab = self.tables['scattering']
        if kind == 'photo':
            stab = self. tables['photoabsorption']

        row = self.query(stab).filter(stab.c.element == elem).all()[0]

        tab_lne = np.array(json.loads(row.log_energy))
        if kind.startswith('coh'):
            tab_val = np.array(json.loads(row.log_coherent_scatter))
            tab_spl = np.array(json.loads(row.log_coherent_scatter_spline))
        elif kind.startswith('incoh'):
            tab_val = np.array(json.loads(row.log_incoherent_scatter))
            tab_spl = np.array(json.loads(row.log_incoherent_scatter_spline))
        else:
            tab_val = np.array(json.loads(row.log_photoabsorption))
            tab_spl = np.array(json.loads(row.log_photoabsorption_spline))

        en = 1.0*as_ndarray(energies)
        emin_tab = 10*int(0.102*np.exp(tab_lne[0]))
        en[np.where(en < emin_tab)] = emin_tab
        out = np.exp(elam_spline(tab_lne, tab_val, tab_spl, np.log(en)))

        if isinstance(energies, (int, float)):
            return out[0]

        return out

    def mu_elam(self, element, energies, kind='total'):
        """
        returns attenuation cross section for an element at energies (in eV)

        Parameters:
            element (string or int):  atomic number or symbol for element
            energies (float or ndarray): energies (in eV) to calculate cross-sections
            kind (string):  one of 'photo' or 'total' for photo-electric or
                  total attenuation, respectively.  Default is 'total'.

        Returns:
           ndarray of scattering values in units of cm^2/gr

        References:
            Elam, Ravel, and Sieber.
        """
        calc = self.cross_section_elam
        kind = kind.lower()
        if kind.startswith('tot'):
            xsec = calc(element, energies, kind='photo')
            xsec += calc(element, energies, kind='coh')
            xsec += calc(element, energies, kind='incoh')
        elif kind.startswith('photo'):
            xsec = calc(element, energies, kind='photo')
        elif kind.lower().startswith('coh'):
            xsec = calc(element, energies, kind='coh')
        elif kind.lower().startswith('incoh'):
            xsec = calc(element, energies, kind='incoh')
        else:
            raise ValueError('unknown cross section kind=%s' % kind)
        return xsec

    def coherent_cross_section_elam(self, element, energies):
        """returns coherenet scattering crossrxr section for an element
        at energies (in eV)

        returns values in units of cm^2 / gr

        arguments
        ---------
        element:  atomic number, atomic symbol for element
        energies: energies in eV to calculate cross-sections

        Data from Elam, Ravel, and Sieber.
        """
        return self.cross_section_elam(element, energies, kind='coh')

    def incoherent_cross_section_elam(self, element, energies):
        """return incoherenet scattering cross section for an element
        at energies (in eV)

        returns values in units of cm^2 / gr

        arguments
        ---------
        element:  atomic number, atomic symbol for element
        energies: energies in eV to calculate cross-sections

        Data from Elam, Ravel, and Sieber.
        """
        return self.cross_section_elam(element, energies, kind='incoh')

    def ionization_potential(self, gas):
        """return effective ionization potential for a gas,
        as appropriate for ionization chambers in the linear
        regime (not in the 'proportional counter' regime)

        returns values in units of eV

        argument
        ---------
        gas (string):  name of gas

        notes
        -----
        Data from G. F. Knoll, Radiation Detection and Measurement, Table 5-1.
        """
        itab = self.tables['ionization_potentials']
        out = self.query(itab).filter(itab.c.gas == gas).all()
        if len(out) != 1:
            raise ValueError('unknown gas for ionization potential: %s' % gas)
        return float(out[0].potential)



'''
This script creates an SQLite3 database of fundamental X-ray fluorescence
parameters as compiled by W.T. Elam, B.D. Ravel and J.R. Sieber, published in
Radiation Physics and Chemistry, 63 (2), 121 (2002). The database is published
by NIST at http://www.cstl.nist.gov/acd/839.01/xrfdownload.html
'''

import io
import json
import os
import time
import sqlite3
from collections import namedtuple

def add_Version(dest, append=True):
    """add Version Information"""
    if os.path.exists(dest) and not append:
        raise IOError('File "%s" already exists -- cannot add Version')

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute('''create table Version (id integer primary key, tag text, date text, notes text)''')

    source = 'Version.dat'
    version_lines = []
    if os.path.exists(source):
        with io.open(source, mode='r', encoding='ascii') as f:
            version_lines = f.readlines()
    rowid = 0
    for line in version_lines:
        if not line.startswith('#') and len(line)> 3:
            _tag, _date, _notes = [l.strip() for l in line.split('//', 2)]
            rowid += 1
            c.execute('insert into Version values (?,?,?,?)',
                      (rowid, _tag, _date, _notes))

    conn.commit()
    c.close()


def add_elementaldata(dest):
    source = 'elemental_data.txt'
    if not os.path.isfile(source):
        raise IOError('File "%s" does not exist' % source)

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute('''create table elements (atomic_number integer primary key,
        element text, molar_mass real, density real)
        ''')
    with io.open(source, encoding='ascii') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            num, sym, mw, rho = line[:-1].split()
            c.execute('insert into elements values (?,?,?,?)', (num, sym, mw, rho))
    conn.commit()
    c.close()

def add_ionization_potentials(dest):
    source = 'ion_chamber_potentials.txt'
    if not os.path.isfile(source):
        raise IOError('File "%s" does not exist' % source)

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute('create table ionization_potentials (gas text,  potential real)')
    with io.open(source, encoding='ascii') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue

            line = line[:-1].strip()
            if len(line)  > 2:
                words = line.split()
                potential = words.pop()
                gas = ' '.join(words)
                c.execute('insert into ionization_potentials values (?,?)', (gas, potential))
    conn.commit()
    c.close()

def add_corehole_data(dest, append=True):
    """add core-widths from Keski-Rahkonen and Krause. Data from
    Atomic Data and Nuclear Data Tables, Vol 14, Number 2, 1974 and
    from  Krause and Oliver, J. Phys. Chem. Ref. Data 8,329 (1979)

    """
    kk_source = 'keskirahkonen_krause.dat'
    ko_source = 'krause_oliver1979.dat'
    if os.path.exists(dest) and not append:
        raise IOError('File "%s" already exists -- cannot add core hole data')

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute(
        '''create table KeskiRahkonen_Krause (id integer primary key,
        atomic_number integer, element text, edge text, width float)''')
    c.execute(
        '''create table Krause_Oliver (id integer primary key,
        atomic_number integer, element text, edge text, width float)''')
    c.execute(
        '''create table corelevel_widths (id integer primary key,
        atomic_number integer, element text, edge text, width float)''')

    # Keski-Rahkonen and Krause data:
    f = open(kk_source)
    lines = f.readlines()
    id = 0
    for line in lines:
        if line.startswith('#'):
            continue
        atno, elem, edge, width = line[:-1].split()
        id +=1
        atno, width = int(atno), float(width)
        c.execute('insert into KeskiRahkonen_Krause values (?,?,?,?,?)',
                      (id, atno, elem, edge, width))
        c.execute('insert into corelevel_widths values (?,?,?,?,?)',
                      (id, atno, elem, edge, width))

    # Krause and Oliver data
    f = open(ko_source)
    lines = f.readlines()
    id = 0
    for line in lines:
        if line.startswith('#'):
            continue
        words = line[:-1].strip().split()
        atno, elem, kwid, l1wid, l2wid, l3wid = words[:6]
        id +=1
        atno = int(atno)
        c.execute('insert into Krause_Oliver values (?,?,?,?,?)',
                      (id, atno, elem, 'K', float(kwid)))
        id +=1
        c.execute('insert into Krause_Oliver values (?,?,?,?,?)',
                      (id, atno, elem, 'L1', float(l1wid)))
        id +=1
        c.execute('insert into Krause_Oliver values (?,?,?,?,?)',
                      (id, atno, elem, 'L2', float(l2wid)))
        id +=1
        c.execute('insert into Krause_Oliver values (?,?,?,?,?)',
                      (id, atno, elem, 'L3', float(l3wid)))

        c.execute('update corelevel_widths set width=? where atomic_number=? and edge=?',
                      (float(kwid), atno, 'K'))
        c.execute('update corelevel_widths set width=? where atomic_number=? and edge=?',
                      (float(l1wid), atno, 'L1'))
        c.execute('update corelevel_widths set width=? where atomic_number=? and edge=?',
                      (float(l2wid), atno, 'L2'))
        c.execute('update corelevel_widths set width=? where atomic_number=? and edge=?',
                      (float(l3wid), atno, 'L3'))

    conn.commit()
    c.close()

def add_Krause_Oliver(dest, append=True):
    """add core-widths from Krause and Oliver,
    J. Phys. Chem. Ref. Data 8,329 (1979)
    """
    source = 'krause_oliver1979.dat'
    if os.path.exists(dest) and not append:
        raise IOError('File "%s" already exists -- cannot add core hole data')

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute(
        '''create table Krause_Oliver (id integer primary key,
        atomic_number integer, element text, edge text, width float)''')


    conn.commit()
    c.close()


def add_Waasmaier(dest, append=True):
    """add f0 data from Waasmaier and Kirfel"""
    source = 'waasmaeir_kirfel.dat'

    if os.path.exists(dest) and not append:
        raise IOError('File "%s" already exists -- cannot add f0 data')

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute(
        '''create table Waasmaier (id integer primary key,
        atomic_number integer, element text, ion text,
        offset real, scale text, exponents text)
        ''')

    f = open(source)
    lines = f.readlines()
    if 'Elastic Photon-Atom Scatt' not in lines[1]:
        raise RuntimeError('Source file not recognized for f0_WaasKirf data')

    strip_ion = str.maketrans('0123456789+-', ' '*12)
    id = 0
    while lines:
        line = lines.pop(0)
        if line.startswith('#S '):
            id += 1
            #print [s for s in line[3:].split()]
            zstr, ion = [s.strip() for s in line[3:].split()]
            atno = int(zstr)
            for i in range(3):
                line = lines.pop(0)
            words = [float(w.strip()) for w in line.split()]
            off   = words[5]
            scale = json.dumps(words[:5])
            expon = json.dumps(words[6:])

            elem = ion.translate(strip_ion).strip()
            for suffix in (('va', 'val')):
                if elem.endswith(suffix):
                    elem = elem[:-len(suffix)]

            c.execute('insert into Waasmaier values (?,?,?,?,?,?,?)',
                      (id, atno, elem, ion, off, scale, expon))

    conn.commit()
    c.close()

def add_Chantler(dest, append=True, table='Chantler', subdir='fine', suffix='.dat'):
    """add f' / f'', mu data from Chantler"""
    dirname = os.path.join('chantler', subdir)

    if os.path.exists(dest) and not append:
        raise IOError('File "%s" already exists -- cannot add f0 data')

    conn = sqlite3.connect(dest)
    c = conn.cursor()
    c.execute(
        '''create table %s (id integer primary key,
        element text, sigma_mu real, mue_f2 real, density real,
        corr_henke float, corr_cl35 float, corr_nucl float,
        energy text, f1 text, f2 text, mu_photo text,
        mu_incoh text, mu_total text)
        ''' % table)

    args = '(%s)' % ','.join(['?']*14)

    nelem = 92
    for z in range(1, nelem+1):
        fname = os.path.join(dirname, '%2.2i.dat' % z)
        lines = open(fname, 'r').readlines()

        # line 1: take symbol and density only
        words = lines[0][1:-1].split()
        words.pop()
        density = float(words.pop())
        elem = words[0].replace(':','')

        # line 2: take sigma_mu
        words = lines[1][1:-1].split()
        sigma_mu = float(words.pop())

        # line 3: take mue_f2
        words = lines[2][1:-1].split()
        mue_f2 = float(words.pop())

        en, f1, f2, mu_photo, mu_incoh, mu_total = [], [], [], [], [], []
        corr_henke, corr_cl35, corr_nucl = 0, 0, 0
        for line in lines:
            if line.startswith('#'):
                if 'Relativistic' in line or 'Nuclear Thomson' in line:
                    line = line[1:-1].replace('#', ' ').strip()
                    label, val = line.split('=')
                    val = val.replace(',', '').replace('e/atom', '')
                    if 'Relativistic' in line:
                        corr_henke, corr_cl35 = [float(x) for x in val.split()]
                    else:
                        corr_nucl = float(val)
                continue

            words = [float(w) for w in line[:-1].split()]
            en.append(1000.0*words[0])
            f1.append(words[1]  - z + corr_cl35 + corr_nucl)
            f2.append(words[2])
            mu_photo.append(words[3])
            mu_incoh.append(words[4])
            mu_total.append(words[3]+words[4])

        query = 'insert into %s values %%s' % table

        c.execute(query % args,
                  (z, elem, sigma_mu, mue_f2, density,
                   corr_henke, corr_cl35, corr_nucl,
                   json.dumps(en), json.dumps(f1), json.dumps(f2),
                   json.dumps(mu_photo), json.dumps(mu_incoh),
                   json.dumps(mu_total)))

    conn.commit()
    c.close()


def add_Elam(dest, overwrite=False, silent=False):
    source = 'elam.dat'
    if not os.path.isfile(source):
        if silent:
            return
        raise IOError('File "%s" does not exist' % source)
    if os.path.isfile(dest) and overwrite:
        os.remove(dest)
    if os.path.exists(dest):
        if silent:
            return
        raise IOError('File "%s" already exists. Use "-f" to overwrite' % dest)

    with io.open(source, encoding='ascii') as f:
        lines = f.readlines()
        if 'Elam, Ravel, Sieber' not in lines[0]:
            raise RuntimeError('Source file not recognized')
        while lines[0].startswith('/'):
            lines.pop(0)

    conn = sqlite3.connect(dest)
    c = conn.cursor()

    current_edge_id = 0
    c.execute(
        '''create table xray_levels (id integer primary key, element text,
        iupac_symbol text, absorption_edge real, fluorescence_yield real,
        jump_ratio real)
        '''
        )
    current_line_id = 0
    c.execute(
        '''create table xray_transitions (id integer primary key, element text,
        iupac_symbol text, siegbahn_symbol text, initial_level text,
        final_level text, emission_energy real, intensity real)
        '''
        )
    current_ck_id = 0
    c.execute(
        '''create table Coster_Kronig (id integer primary key, element text,
        initial_level text, final_level text,
        transition_probability real, total_transition_probability real)
        '''
        )
    current_photo_id = 0
    c.execute(
        '''create table photoabsorption (id integer primary key, element text,
        log_energy text, log_photoabsorption text,
        log_photoabsorption_spline text)
        '''
        )
    current_scatter_id = 0
    c.execute(
        '''create table scattering (id integer primary key, element text,
        log_energy text,
        log_coherent_scatter text, log_coherent_scatter_spline text,
        log_incoherent_scatter text, log_incoherent_scatter_spline text)
        '''
        )

    while lines:
        line = lines.pop(0)
        if line.startswith('Element'):
            sym, num, mw, rho = line.split()[1:]
            current_element = sym
        elif line.startswith('Edge'):
            current_edge_id += 1
            label, energy, yield_, jump = line.split()[1:]
            el = current_element
            c.execute(
                'insert into xray_levels values (?,?,?,?,?,?)',
                (current_edge_id, el, label, energy, yield_, jump)
                )
            current_edge = label
        elif line.startswith('  Lines'):
            while True:
                if lines[0].startswith('    '):
                    current_line_id += 1
                    line = lines.pop(0)
                    iupac, siegbahn, energy, intensity = line.split()
                    start, end = iupac.split('-')
                    el = current_element
                    c.execute(
                        'insert into xray_transitions values (?,?,?,?,?,?,?,?)',
                        (current_line_id, el, iupac, siegbahn, start, end,
                        energy, intensity)
                        )
                else:
                    break
        elif line.startswith('  CK '):
            temp = line.split()[1:]
            ck = []
            while temp:
                (i,j), temp = temp[:2], temp[2:]
                ck.append((i,j))
            if lines[0].startswith('  CKtotal'):
                temp = lines.pop(0).split()[1:]
                ck_total = []
                while temp:
                    (i,j), temp = temp[:2], temp[2:]
                    ck_total.append((i,j))
            else:
                ck_total = ck
            for i, j in zip(ck, ck_total):
                current_ck_id += 1
                (so, p), tp = i[:], j[1]
                c.execute(
                    '''insert into Coster_Kronig
                    values (?,?,?,?,?,?)''',
                    (current_ck_id, current_element, current_edge, so, p, tp)
                    )
        elif line.startswith('Photo'):
            current_photo_id += 1
            energy = []
            photo = []
            spline = []
            while lines[0].startswith('    '):
                temp = [float(i) for i in lines.pop(0).split()]
                energy.append(temp[0])
                photo.append(temp[1])
                spline.append(temp[2])
            c.execute(
                'insert into photoabsorption values (?,?,?,?,?)',
                (current_photo_id, current_element, json.dumps(energy),
                json.dumps(photo), json.dumps(spline))
                )
        elif line.startswith('Scatter'):
            current_scatter_id += 1
            energy = []
            cs = []
            css = []
            ics = []
            icss = []
            while lines[0].startswith('    '):
                temp = [float(i) for i in lines.pop(0).split()]
                energy.append(temp[0])
                cs.append(temp[1])
                css.append(temp[2])
                ics.append(temp[3])
                icss.append(temp[4])
            c.execute(
                'insert into scattering values (?,?,?,?,?,?,?)',
                (current_scatter_id, current_element, json.dumps(energy),
                json.dumps(cs), json.dumps(css), json.dumps(ics),
                json.dumps(icss))
                )

    conn.commit()
    c.close()

_EADL_doc = """
The 1997 release of the Evaluated Atomic Data Library (EADL97)

This module parses the EADL.DAT file that can be downloaded from:

http://www-nds.iaea.org/epdl97/libsall.htm

EADL contains atomic relaxation information for use in particle transport
analysis for atomic number Z = 1-100 and for each subshell.

The original units are in cm and MeV.

The specific data are:

- Subshell data

    a) number of electrons
    b) binding and kinetic energy (MeV)
    c) average radius (cm)
    d) radiative and non-radiative level widths (MeV)
    e) average number of released electrons and x-rays
    f) average energy of released electrons and x-rays (MeV)
    g) average energy to the residual atom, i.e., local deposition (MeV)

- Transition probability data

    a) radiation transition probabilities
    b) non-radiative transition probabilities

The data are organized in blocks with headers.

The first line of the header:

Columns    Format   Definition
1-3         I3      Z  - atomic number
4-6         I3      A  - mass number (in all cases=0 for elemental data)
8-9         I2      Yi - incident particle designator (7 is photon)
11-12       I2      Yo - outgoing particle designator (0, no particle
                                                       7, photon
                                                       8, positron
                                                       9, electron)
14-24       E11.4   AW - atomic mass (amu)

26-31       I6      Date of evaluation (YYMMDD)

The second line of the header:

Columns    Format   Definition
1-2         I2      C  - reaction descriptor
                                  = 71, coherent scattering
                                  = 72, incoherent scattering
                                  = 73, photoelectric effect
                                  = 74, pair production
                                  = 75, triplet production
                                  = 91, subshell parameters
                                  = 92, transition probabilities
                                  = 93, whole atom parameters

3-5         I2      I  - reaction property:
                                  =   0, integrated cross section
                                  =  10, avg. energy of Yo
                                  =  11, avg. energy to the residual atom
                                  = 912, number of electrons
                                  = 913, binding energy
                                  = 914, kinetic energy
                                  = 915, average radius
                                  = 921, radiative level width
                                  = 922, non-radiative level width
                                  = 931, radiative transition probability
                                  = 932, non-radiative transition probability
                                  = 933, particles per initial vacancy
                                  = 934, energy of particles per initial vacancy
                                  = 935, average energy to the residual atom, i.e.
                                         local deposition, per initial vacancy
                                  --- moved to EPDL97 ---
                                  = 941, form factor
                                  = 942, scattering function
                                  = 943, imaginary anomalous scatt. factor
                                  = 944, real anomalous scatt. factor

6-8         I3      S  - reaction modifier:
                                  =  0 no X1 field data required
                                  = 91 X1 field data required

22-32       #11.4   X1 - subshell designator
                                      0 if S is 0
                                      if S is 91, subshell designator


                 Summary of the EADL Data Base
--------------------------------------------------------------------------
Yi    C    S    X1    Yo   I          Data Types
--------------------------------------------------------------------------
                     Subshell parameters
--------------------------------------------------------------------------
0    91    0    0.    0    912        number of electrons
0    91    0    0.    0    913        binding energy
0    91    0    0.    0    914        kinetic energy
0    91    0    0.    0    915        average radius
0    91    0    0.    0    921        radiative level width
0    91    0    0.    0    921        non-radiative level width
--------------------------------------------------------------------------
                     Transititon probabilities
--------------------------------------------------------------------------
0    92    0    0.    0    935        average energy to the residual atom
0    92    0    0.  7 or 9 933        average number of particles per
                                      initial vacancy
0    92    0    0.  7 or 9 934        average energy of particles per
                                      initial vacancy
0    92   91    *     0    931        radiative transition probability
0    92   91    *     0    932        non-radiative transition probability
---------------------------------------------------------------------------
Yi    C    S    X1    Yo   I          Data Types
--------------------------------------------------------------------------

* -> Subshell designator

Data sorted in ascending order Z -> C -> S -> X1 -> Yo -> I
"""


def parse_EADL(fname):
    '''Parse the EADL data file

    Data source:

    http://www-nds.iaea.org/epdl97/libsall.htm

    Both returned dictionaries share the same keys (which are a subset of
    the header data.

    The values in the key are ``['Z', 'C', 'S', 'X1', 'Yo', 'I']``

      Z  : atomic number
      C  : {91, 92} <-> {subshell, transition}
      S  : if section depends on shell
      X1 : shell if S
      Yo : particle out {7, 9, 0} <-> {photon, electron, none}
      I  : property key
      Yi : incoming particle {0, 7} <-> {none, photon}

    Parameters
    ----------
    fname : str

    Returns
    -------
    headers : dict
        All relevant header information with human-readable aliases.

    data : dict
        Lists of namedtuple instances for this sub-table


    ''' + _EADL_doc
    SHELL_MAP = {1: 'K (1s1/2)',
                 2: 'L (2)',
                 3: 'L1 (2s1/2)',
                 4: 'L23 (2p)',
                 5: 'L2 (2p1/2)',
                 6: 'L3 (2p3/2)',
                 7: 'M (3)',
                 8: 'M1 (3s1/2)',
                 9: 'M23 (3p)',
                 10: 'M2 (3p1/2)',
                 11: 'M3 (3p3/2)',
                 12: 'M45 (3d)',
                 13: 'M4 (3d3/2)',
                 14: 'M5 (3d5/2)',
                 15: 'N (4)',
                 16: 'N1 (4s1/2)',
                 17: 'N23 (4p)',
                 18: 'N2 (4p1/2)',
                 19: 'N3 (4p3/2)',
                 20: 'N45 (4d)',
                 21: 'N4 (4d3/2)',
                 22: 'N5 (4d5/2)',
                 23: 'N67 (4f)',
                 24: 'N6 (4f5/2)',
                 25: 'N7 (4f7/2)',
                 26: 'O (5)',
                 27: 'O1 (5s1/2)',
                 28: 'O23 (5p)',
                 29: 'O2 (5p1/2)',
                 30: 'O3 (5p3/2)',
                 31: 'O45 (5d)',
                 32: 'O4 (5d3/2)',
                 33: 'O5 (5d5/2)',
                 34: 'O67 (5f)',
                 35: 'O6 (5f5/2)',
                 36: 'O7 (5f7/2)',
                 37: 'O89 (5g)',
                 38: 'O8 (5g7/2)',
                 39: 'O9 (5g9/2)',
                 40: 'P (6)',
                 41: 'P1 (6s1/2)',
                 42: 'P23 (6p)',
                 43: 'P2 (6p1/2)',
                 44: 'P3 (6p3/2)',
                 45: 'P45 (6d)',
                 46: 'P4 (6d3/2)',
                 47: 'P5 (6d5/2)',
                 48: 'P67 (6f)',
                 49: 'P6 (6f5/2)',
                 50: 'P7 (6f7/2)',
                 51: 'P89 (6g)',
                 52: 'P8 (6g7/2)',
                 53: 'P9 (6g9/2)',
                 54: 'P1011 (6h)',
                 55: 'P10 (6h9/2)',
                 56: 'P11 (6h11/2)',
                 57: 'Q (7)',
                 58: 'Q1 (7s1/2)',
                 59: 'Q23 (7p)',
                 60: 'Q2 (7p1/2)',
                 61: 'Q3 (7p3/2)'}

    Elements = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
                'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca',
                'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
                'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr',
                'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In',
                'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr',
                'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
                'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt',
                'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr',
                'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',
                'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg',
                'Bh', 'Hs', 'Mt']

    KeyTuple = namedtuple('KeyTuple', ['Z', 'C', 'S', 'X1', 'Yo', 'I', 'Yi'])

    def make_dataline_key(inp_dict):
        return KeyTuple(*tuple(inp_dict[k] for k in KeyTuple._fields))

    reaction_proprety_map = {
        0: 'cross_section',
        10: 'secondary_particle_energy',
        11: 'atom_energy_transfer',
        912: 'number_of_electrons',
        913: 'binding_energy',
        914: 'kinetic_energy',
        915: 'average_radius',
        921: 'radiative_level_width',
        922: 'non-radiative_level_width',
        931: 'radiative_transition_probability',
        932: 'non-radiative_transition_probability',
        933: 'particles_per_initial_vacancy',
        934: 'energy_of_particles_per_initial_vacancy',
        935: 'average_energy_to_the_residual_atom',
        941: 'form_factor',
        942: 'scattering_function',
        943: 'imaginary_anomalous_scattering_factor',
        944: 'real_anomalous_scattering_factor',
    }
    particle_map = {7: 'photon', 0: 'none', 9: 'electron', 8: 'positron'}
    reaction_code_map = {91: 'subshell', 92: 'transition',
                         71: 'coherent scattering',
                         72: 'incoherent scattering',
                         73: 'photoelectric effect',
                         74: 'pair production',
                         75: 'triplet production',
                         93: 'whole atom parameters'}
    interpolation_map = {0: 'linear in x and y',
                         2: 'linear in x and y',
                         3: 'logarithmic in x, linear in y',
                         4: 'linear in x, logarithmic in y',
                         5: 'logarithmic in x and y'}

    def _proc_shell_int(header, row):
        klass = reaction_proprety_classes[header['I']]
        i, N = map(int, row)
        return klass(SHELL_MAP[i], i, N)

    def _proc_shell_float(header, row):
        klass = reaction_proprety_classes[header['I']]
        i, f = row
        i = int(i)
        return klass(SHELL_MAP[i], i, f)

    def _proc_radiative_transfer(header, row):
        klass = reaction_proprety_classes[header['I']]
        i = int(header['X1'])
        j, fr, Er = row
        j = int(j)
        return klass(SHELL_MAP[i], i, SHELL_MAP[j], j, fr, Er)

    def _proc_nonradiative_transfer(header, row):
        klass = reaction_proprety_classes[header['I']]
        i = int(header['X1'])
        j, k, fnr, Enr = row
        j = int(j)
        k = int(k)
        return klass(SHELL_MAP[i], i,
                     SHELL_MAP[j], j,
                     SHELL_MAP[k], k,
                     fnr, Enr)

    def _proc_float_float(header, row):
        klass = reaction_proprety_classes[header['I']]
        return klass(*row)

    reaction_proprety_classes = {
        0: namedtuple('IntegratedCrossSection',
                      ('E_incident', 'cross_section')),
        10: namedtuple('AverageEneregySecondary',
                       ('E_incident', 'E_secondary')),
        11: namedtuple('AverageEnergyResidual', ('E_incident', 'E_residual')),
        912: namedtuple('NumberOfElectrons', ('shell', 'shell_code', 'N')),
        913: namedtuple('BindingEnergy', ('shell', 'shell_code', 'E_be')),
        914: namedtuple('KeneticEnergy', ('shell', 'shell_code', 'E_ke')),
        915: namedtuple('AverageRadius', ('shell', 'shell_code', 'r_mean')),
        921: namedtuple('RadiativeLevelWidth',
                        ('shell', 'shell_code', 'gamma_r')),
        922: namedtuple('NonRadiativeLevelWidth',
                        ('shell', 'shell_code', 'gamma_nr')),
        931: namedtuple('RadiativeTransitionProbability',
                        ('primary_shell', 'primary_shell_code',
                         'secondary_shell', 'secondary_shell_code',
                         'transition_probability', 'E')),
        932: namedtuple('NonRadiativeTransitionProbability',
                        ('primary_shell', 'primary_shell_code',
                         'secondary_shell', 'secondary_shell_code',
                         'tertiary_shell', 'tertiary_shell_code',
                         'transition_probability', 'E')),
        933: namedtuple('ParticlesPerInitVacency',
                        ('shell', 'shell_code', 'N_p')),
        934: namedtuple('EnergePerInitVacency',
                        ('shell', 'shell_code', 'E_p')),
        935: namedtuple('AverageEofRisdualAtom',
                        ('shell', 'shell_code', 'E_mean')),
        941: namedtuple('FormFactor', ('x', 'F')),
        942: namedtuple('ScatteringFunction', ('x', 'S')),
        943: namedtuple('ImAnaomalousScatteringFactor', ('E_incident', 'Im')),
        944: namedtuple('ReAnaomalousScatteringFactor', ('E_incident', 'Re')),
        }

    reaction_property_funcs = {
        0: _proc_float_float,
        10: _proc_float_float,
        11: _proc_float_float,
        912: _proc_shell_int,
        913: _proc_shell_float,
        914: _proc_shell_float,
        915: _proc_shell_float,
        921: _proc_shell_float,
        922: _proc_shell_float,
        931: _proc_radiative_transfer,
        932: _proc_nonradiative_transfer,
        933: _proc_shell_float,
        934: _proc_shell_float,
        935: _proc_shell_float,
        941: _proc_float_float,
        942: _proc_float_float,
        943: _proc_float_float,
        944: _proc_float_float,
        }

    BREAK_TOKEN = ' ' * 71 + '1'

    def _fixed_width_float(val):
        split = 8 if val[8] in {'+', '-'} else 9
        base = float(val[:split])
        exp = val[split:].replace(' ', '')
        if exp:
            exp = int(exp)
        else:
            exp = 0
        return base * 10 ** exp

    in_section = False
    expect_second_header_line = False
    ret_header = {}
    ret_data = {}
    current_key = None
    cur_header = None
    with open(fname, 'r') as fin:

        for ln in fin:
            ln = ln.rstrip('\n\r')
            if expect_second_header_line:
                cur_header['C'] = C = int(ln[0:2])
                cur_header['I'] = I = int(ln[2:5])
                cur_header['S'] = S = int(ln[5:8])
                cur_header['X1'] = X1 = int(_fixed_width_float(ln[21:32]))
                cur_header['reaction_code'] = reaction_code_map[C]
                cur_header['reaction_property'] = reaction_proprety_map[I]
                cur_header['reaction_property_code'] = I
                if S == 91:
                    cur_header['subshell_code'] = X1
                    if X1 != 0:
                        cur_header['subshell'] = SHELL_MAP[X1]
                    else:
                        cur_header['subshell'] = 'none'
                elif S == 0 and X1 == 0:
                    cur_header['subshell_code'] = 0
                    cur_header['subshell'] = 'none'
                else:
                    raise ValueError('Inconsistent Data X1 = {!r}, '
                                     'S = {!r}'.forma(X1, S))

                expect_second_header_line = False
                key = make_dataline_key(cur_header)
                ret_header[key] = cur_header
                ret_data[key] = []
                current_key = key
                in_section = True
            elif not in_section:
                # read the first line
                expect_second_header_line = True

                cur_header = dict()
                cur_header['Z'] = Z = int(ln[0:3])
                cur_header['A'] = A = int(ln[3:6])
                cur_header['Yi'] = Yi = int(ln[7:9])
                cur_header['Yo'] = Yo = int(ln[10:12])
                cur_header['Aw'] = Aw = _fixed_width_float(ln[13:24])
                Iflag = ln[31]
                if Iflag == ' ':
                    Iflag = 0
                else:
                    Iflag = int(Iflag)
                cur_header['Iflag'] = Iflag

                cur_header['element'] = Elements[Z-1]
                cur_header['atomic_number'] = Z
                cur_header['mass_number'] = A
                cur_header['atomic_mass'] = Aw
                cur_header['incoming_particle'] = particle_map[Yi]
                cur_header['incoming_particle_value'] = Yi
                cur_header['outgoing_particle'] = particle_map[Yo]
                cur_header['outgoing_particle_value'] = Yo
                cur_header['interpolation'] = interpolation_map[Iflag]
            elif ln == BREAK_TOKEN:
                in_section = False
            else:
                # get function to process rows of this type
                proc_func = reaction_property_funcs.get(
                    cur_header['I'], lambda h, row: list(row))
                # parse the fixed with data to floats
                row = [_fixed_width_float(ln[j*11:(j+1)*11])
                       for j in range(len(ln) // 11)]
                # generate final result representation
                ret_data[current_key].append(proc_func(cur_header, row))
        return ret_header, ret_data

if __name__ == '__main__':
    try:
        import argparse
    except ImportError:
        raise RuntimeError(
            'argparse module not found.\n'
            'Install argparse or update to python-2.7 or >=python-3.2.'
            )
    parser = argparse.ArgumentParser(
        description='export the Elam, Waasmaier, Chantler data to an SQLite database "dest"'
        )
    dest = 'xraydb.sqlite'
    parser.add_argument('-f', '--force', action='store_true')
    parser.add_argument('-s', '--silent', action='store_true')
    args = parser.parse_args()

    add_Elam(dest, overwrite=args.force, silent=args.silent)
    add_Waasmaier(dest, append=True)
    add_elementaldata(dest)
    add_ionization_potentials(dest)
    add_corehole_data(dest, append=True)
    add_Chantler(dest, table='Chantler',        subdir='fine',   append=True)
    add_Version(dest)



import os
import numpy as np
from collections import namedtuple
from .chemparser import chemparse
from .xray import mu_elam, atomic_mass
from .utils import get_homedir

_materials = None

Material = namedtuple('Material', ('formula', 'density', 'name', 'categories'))

def get_user_materialsfile():
    """return name for user-specific materials.dat file
    With $HOME being the users home directory, this will be
    $HOME/.config/xraydb/materials.dat
    """
    return os.path.join(get_homedir(), '.config', 'xraydb', 'materials.dat')

def _read_materials_db():
    """return _materials dictionary, creating it if needed"""
    global _materials
    if _materials is None:
        # initialize materials table
        _materials = {}

        def read_materialsfile(fname):
            with open(fname, 'r') as fh:
                lines = fh.readlines()
            for line in lines:
                line = line.strip()
                if len(line) > 2 and not line.startswith('#'):
                    words = [i.strip() for i in line.split('|')]
                    name = words[0].lower()
                    formula = None
                    if len(words) == 3: # older style
                        # "name | formula | density"  or  "name | density | formula"
                        iformula = 1
                        try:
                            density = float(words[2])
                        except ValueError:
                            density = float(words[1])
                            iformula = 2
                        formula = words[iformula]
                        categories = []
                    elif len(words) == 4: # newer style, with categories
                        density = float(words[1])
                        categories = [w.strip() for w in words[2].split(',')]
                        formula = words[3]
                    if formula is not None:
                        formula = formula.replace(' ', '')
                        _materials[name] = Material(formula, density, name, categories)

        # first, read from standard list
        local_dir, _ = os.path.split(__file__)
        fname = os.path.join(local_dir, 'materials.dat')
        if os.path.exists(fname):
            read_materialsfile(fname)
        # next, read from users materials file
        fname = get_user_materialsfile()
        if os.path.exists(fname):
            read_materialsfile(fname)

    return _materials

def material_mu(name, energy, density=None, kind='total'):
    """X-ray attenuation length (in 1/cm) for a material by name or formula

    Args:
        name (str): chemical formul or name of material from materials list.
        energy (float or ndarray):   energy or array of energies in eV
        density (None or float):  material density (gr/cm^3).
        kind (str): 'photo' or 'total' for whether to return the
                    photo-absorption or total cross-section ['total']
    Returns:
        absorption length in 1/cm

    Notes:
        1.  material names are not case sensitive,
            chemical compounds are case sensitive.
        2.  mu_elam() is used for mu calculation.
        3.  if density is None and material is known, that density will be used.

    Examples:
        >>> material_mu('H2O', 10000.0)
        5.32986401658495
    """
    global _materials
    if _materials is None:
        _materials = _read_materials_db()
    formula = None
    _density  = None
    mater = _materials.get(name.lower(), None)
    if mater is None:
        for key, val in _materials.items():
            if name.lower() == val[0].lower(): # match formula
                mater = val
                break

    # default to using passed in name as a formula
    if formula is None:
        if mater is None:
            formula = name
        else:
            formula = mater.formula
    if density is None and mater is not None:
        density = mater.density
    if density is None:
        raise Warning('material_mu(): must give density for unknown materials')

    mass_tot, mu = 0.0, 0.0
    for elem, frac in chemparse(formula).items():
        mass  = frac * atomic_mass(elem)
        mu   += mass * mu_elam(elem, energy, kind=kind)
        mass_tot += mass
    return density*mu/mass_tot


def material_mu_components(name, energy, density=None, kind='total'):
    """material_mu_components: absorption coefficient (in 1/cm) for a compound

    Args:
        name (str): chemical formul or name of material from materials list.
        energy (float or ndarray):   energy or array of energies in eV
        density (None or float):  material density (gr/cm^3).
        kind (str):  'photo' or 'total'for whether to
                  return photo-absorption or total cross-section ['total']

    Returns:
        dict for constructing mu per element,
        with elements 'mass' (total mass), 'density', and
       'elements' (list of atomic symbols for elements in material).
        For each element, there will be an item (atomic symbol as key)
        with tuple of (stoichiometric fraction, atomic mass, mu)

    Examples:
        >>> xraydb.material_mu('quartz', 10000)
        50.36774553547068
        >>> xraydb.material_mu_components('quartz', 10000)
        {'mass': 60.0843, 'density': 2.65, 'elements': ['Si', 'O'],
        'Si': (1, 28.0855, 33.87943243018506), 'O': (2.0, 15.9994, 5.952824815297084)}
     """
    global _materials
    if _materials is None:
        _materials = _read_materials_db()
    mater = _materials.get(name.lower(), None)
    if mater is None:
        formula = name
        if density is None:
            raise Warning('material_mu(): must give density for unknown materials')
    else:
        formula = mater.formula
        density = mater.density

    out = {'mass': 0.0, 'density': density, 'elements':[]}
    for atom, frac in chemparse(formula).items():
        mass  = atomic_mass(atom)
        mu    = mu_elam(atom, energy, kind=kind)
        out['mass'] += frac*mass
        out[atom] = (frac, mass, mu)
        out['elements'].append(atom)
    return out


def get_material(name):
    """look up material name, return formula and density


    Args:
        name (str): name of material or chemical formula

    Returns:
        chemical formula, density of material

    Examples:
        >>> xraydb.get_material('kapton')
        ('C22H10N2O5', 1.43)

    See Also:
       find_material()

    """
    material = find_material(name)
    if material is None:
        return None
    return material.formula, material.density

def find_material(name):
    """look up material name, return material instance

    Args:
        name (str): name of material or chemical formula

    Returns:
        material instance

    Examples:
        >>> xraydb.find_material('kapton')
        Material(formula='C22H10N2O5', density=1.42, name='kapton', categories=['polymer'])

    See Also:
       get_material()

    """
    global _materials
    if _materials is None:
        _materials = _read_materials_db()

    mat =  _materials.get(name.lower(), None)

    if mat is not None:
        return mat
    for mat in _materials.values():
        if mat.formula == name:
            return mat
    return None



def get_materials(force_read=False, categories=None):
    """get dictionary of all available materials

    Args:
        force_read (bool): whether to force a re-reading of the
                         materials database [False]
        categories (list of strings or None): restrict results
                         to those that match category names

    Returns:
        dict with keys of material name and values of Materials instances

    Examples:
        >>> for name, m in xraydb.get_materials().items():
        ...      print(name, m)
        ...
        water H2O 1.0
        lead Pb 11.34
        aluminum Al 2.7
        kapton C22H10N2O5 1.42
        polyimide C22H10N2O5 1.42
        nitrogen N 0.00125
        argon Ar 0.001784
        ...


    """
    global _materials

    if force_read or _materials is None:
        _materials = _read_materials_db()
    if categories is not None:
        if not type(categories) is list:
            categories = list([categories])
        filtered_materials = {
            k: v for k,v in _materials.items() if
            (set(v.categories) & set(categories))
        }
        return filtered_materials
    else:    
        return _materials


def add_material(name, formula, density, categories=None):
    """add a material to the users local material database

    Args:
        name (str): name of material
        formula (str): chemical formula
        density (float): density
        categories (list of strings or None): list of category names

    Returns:
        None

    Notes:
        the data will be saved to $HOME/.config/xraydb/materials.dat
        in the users home directory, and will be useful in subsequent sessions.

    Examples:
        >>> xraydb.add_material('becopper', 'Cu0.98e0.02', 8.3, categories=['metal'])

    """
    global _materials
    if _materials is None:
        _materials = _read_materials_db()
    formula = formula.replace(' ', '')

    if categories is None:
        categories = []
    _materials[name.lower()] = Material(formula, float(density), name, categories)

    fname = get_user_materialsfile()

    if os.path.exists(fname):
        fh = open(fname, 'r')
        text = fh.readlines()
        fh.close()
    else:
        parent, _ = os.path.split(fname)
        if not os.path.exists(parent):
            try:
                os.makedirs(parent)
            except FileExistsError:
                pass
        text = ['# user-specific database of materials\n',
                '# name  |  density |  categories | formulan']

    catstring = ', '.join(categories)
    text.append(" %s | %g  | %s | %s\n" % (name, density, catstring, formula))

    with open(fname, 'w') as fh:
        fh.write(''.join(text))



#!/usr/bin/env python
import os
import sys
import shutil
from setuptools import setup
import versioneer

dbfile = 'xraydb.sqlite'
matfile = 'materials.dat'

dest_dbfile = os.path.join('xraydb', dbfile)
src_dbfile = os.path.join('..', dbfile)
if (not os.path.exists(dest_dbfile) and
    os.path.exists(src_dbfile)):
    shutil.copy(src_dbfile, dest_dbfile)

desc="X-ray Reference Data for the Elements using SQLite"

setup(name = 'xraydb',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      author='Matthew Newville',
      author_email='newville@cars.uchicago.edu',
      url='http://github.com/xraypy/XrayDB',
      download_url='http://github.com/xrapy/XrayDB',
      license='public domain',
      description=desc,
      long_description=desc,
      platforms=('Windows', 'Linux', 'Mac OS X'),
      classifiers=['Intended Audience :: Science/Research',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Scientific/Engineering'],
      install_requires=('sqlalchemy', 'numpy', 'scipy'),
      package_dir={'xraydb': 'xraydb'},
      packages=['xraydb'],
      package_data={'xraydb': [dbfile, matfile]})



#
# xraydb documentation build configuration file

import sys, os
import xraydb

sys.path.insert(0, '.')

import authorlist_format

release = xraydb.__version__.replace('-', ' ').replace('+', ' ')
version = release = release.split(' ')[0]

project = 'xraydb'
html_title = html_short_title = 'X-ray DB:  X-ray Reference Data in SQLite'

author = 'Matthew Newville'
copyright = 'Public Domain'

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.mathjax',
              'sphinx.ext.extlinks', 'sphinxcontrib.napoleon',
              'sphinxcontrib.bibtex' ]

todo_include_todos = True

templates_path = ['_templates']
source_suffix = '.rst'
source_encoding = 'utf-8'
master_doc = 'index'
today_fmt = '%Y-%B-%d'

exclude_trees = ['_build']

add_function_parentheses = True
add_module_names = False
pygments_style = 'sphinx'

bibtex_bibfiles = ['xraydb.bib']

# html themes: 'default', 'sphinxdoc',  'alabaster', 'agogo', 'nature'
html_theme = 'nature'

# html_theme_options = {'pagewidth':'85em',
#                       'documentwidth':'60em',
#                       'sidebarwidth': '25em',
#                       'headercolor1': '#000080',
#                       'headercolor2': '#0000A0',
#                       }

#html_logo = None
#html_favicon = None
html_static_path = ['_static']
html_last_updated_fmt = '%Y-%B-%d'
html_show_sourcelink = True
htmlhelp_basename = 'xraydbdoc'



import sys
from collections import namedtuple
import numpy as np

from .utils import (R_ELECTRON_CM, AVOGADRO, PLANCK_HC, E_MASS,
                    QCHARGE, SI_PREFIXES, index_nearest)

R0 = 1.e8 * R_ELECTRON_CM

from .xraydb import XrayDB,  XrayLine
from .chemparser import chemparse

fluxes = namedtuple('IonChamberFluxes', ('incident', 'transmitted',
                                         'photo', 'incoherent'))

DarwinWidth = namedtuple('DarwinWidth', ('theta', 'theta_offset',
                                         'theta_width', 'theta_fwhm',
                                         'energy_width', 'energy_fwhm',
                                         'zeta', 'dtheta', 'denergy',
                                         'intensity'))

TransmissionSample = namedtuple('TransmissionSample', ('energy_eV',
                                                       'absorp_total',
                                                       'mass_fractions',
                                                       'absorbance_steps',
                                                       'area_cm2',
                                                       'mass_total_mg',
                                                       'mass_components_mg',
                                                       'density',
                                                       'thickness_mm',
                                                       'absorption_length_um'))


_edge_energies = {'k': np.array([-1.0, 13.6, 24.6, 54.7, 111.5, 188.0,
                                 284.2, 409.9, 543.1, 696.7, 870.2, 1070.8,
                                 1303.0, 1559.0, 1839.0, 2145.5, 2472.0,
                                 2822.0, 3205.9, 3608.4, 4038.5, 4492.0,
                                 4966.0, 5465.0, 5989.0, 6539.0, 7112.0,
                                 7709.0, 8333.0, 8979.0, 9659.0, 10367.0,
                                 11103.0, 11867.0, 12658.0, 13474.0,
                                 14326.0, 15200.0, 16105.0, 17038.0,
                                 17998.0, 18986.0, 20000.0, 21044.0,
                                 22117.0, 23220.0, 24350.0, 25514.0,
                                 26711.0, 27940.0, 29200.0, 30491.0,
                                 31814.0, 33169.0, 34561.0, 35985.0,
                                 37441.0, 38925.0, 40443.0, 41991.0,
                                 43569.0, 45184.0, 46834.0, 48519.0,
                                 50239.0, 51996.0, 53789.0, 55618.0,
                                 57486.0, 59390.0, 61332.0, 63314.0,
                                 65351.0, 67416.0, 69525.0, 71676.0,
                                 73871.0, 76111.0, 78395.0, 80725.0,
                                 83102.0, 85530.0, 88005.0, 90526.0,
                                 93105.0, 95730.0, 98404.0, 101137.0,
                                 103922.0, 106755.0, 109651.0, 112601.0,
                                 115606.0, 118669.0, 121791.0, 124982.0,
                                 128241.0, 131556.0]),

                  'l3': np.array([-1.0, -1.0, -1.0, -1.0, 3.0, 4.7, 7.2,
                                  17.5, 18.2, 19.9, 21.6, 30.5, 49.2, 72.5,
                                  99.2, 135.0, 162.5, 200.0, 248.4, 294.6,
                                  346.2, 398.7, 453.8, 512.1, 574.1, 638.7,
                                  706.8, 778.1, 852.7, 932.7, 1021.8,
                                  1116.4, 1217.0, 1323.6, 1433.9, 1550.0,
                                  1678.4, 1804.0, 1940.0, 2080.0, 2223.0,
                                  2371.0, 2520.0, 2677.0, 2838.0, 3004.0,
                                  3173.0, 3351.0, 3538.0, 3730.0, 3929.0,
                                  4132.0, 4341.0, 4557.0, 4786.0, 5012.0,
                                  5247.0, 5483.0, 5723.0, 5964.0, 6208.0,
                                  6459.0, 6716.0, 6977.0, 7243.0, 7514.0,
                                  7790.0, 8071.0, 8358.0, 8648.0, 8944.0,
                                  9244.0, 9561.0, 9881.0, 10207.0, 10535.0,
                                  10871.0, 11215.0, 11564.0, 11919.0,
                                  12284.0, 12658.0, 13035.0, 13419.0,
                                  13814.0, 14214.0, 14619.0, 15031.0,
                                  15444.0, 15871.0, 16300.0, 16733.0,
                                  17166.0, 17610.0, 18057.0, 18510.0,
                                  18970.0, 19435.0]),

                  'l2': np.array([-1.0, -1.0, -1.0, -1.0, 3.0, 4.7, 7.2,
                                  17.5, 18.2, 19.9, 21.7, 30.4, 49.6, 72.9,
                                  99.8, 136.0, 163.6, 202.0, 250.6, 297.3,
                                  349.7, 403.6, 460.2, 519.8, 583.8, 649.9,
                                  719.9, 793.2, 870.0, 952.3, 1044.9,
                                  1143.2, 1248.1, 1359.1, 1474.3, 1596.0,
                                  1730.9, 1864.0, 2007.0, 2156.0, 2307.0,
                                  2465.0, 2625.0, 2793.0, 2967.0, 3146.0,
                                  3330.0, 3524.0, 3727.0, 3938.0, 4156.0,
                                  4380.0, 4612.0, 4852.0, 5107.0, 5359.0,
                                  5624.0, 5891.0, 6164.0, 6440.0, 6722.0,
                                  7013.0, 7312.0, 7617.0, 7930.0, 8252.0,
                                  8581.0, 8918.0, 9264.0, 9617.0, 9978.0,
                                  10349.0, 10739.0, 11136.0, 11544.0,
                                  11959.0, 12385.0, 12824.0, 13273.0,
                                  13734.0, 14209.0, 14698.0, 15200.0,
                                  15711.0, 16244.0, 16785.0, 17337.0,
                                  17907.0, 18484.0, 19083.0, 19693.0,
                                  20314.0, 20948.0, 21600.0, 22266.0,
                                  22952.0, 23651.0, 24371.0]),


                  'l1': np.array([-1.0, -1.0, -1.0, 5.3, 8.0, 12.6, 18.0,
                                  37.3, 41.6, 45.0, 48.5, 63.5, 88.6,
                                  117.8, 149.7, 189.0, 230.9, 270.0, 326.3,
                                  378.6, 438.4, 498.0, 560.9, 626.7, 696.0,
                                  769.1, 844.6, 925.1, 1008.6, 1096.7,
                                  1196.2, 1299.0, 1414.6, 1527.0, 1652.0,
                                  1782.0, 1921.0, 2065.0, 2216.0, 2373.0,
                                  2532.0, 2698.0, 2866.0, 3043.0, 3224.0,
                                  3412.0, 3604.0, 3806.0, 4018.0, 4238.0,
                                  4465.0, 4698.0, 4939.0, 5188.0, 5453.0,
                                  5714.0, 5989.0, 6266.0, 6548.0, 6835.0,
                                  7126.0, 7428.0, 7737.0, 8052.0, 8376.0,
                                  8708.0, 9046.0, 9394.0, 9751.0, 10116.0,
                                  10486.0, 10870.0, 11271.0, 11682.0,
                                  12100.0, 12527.0, 12968.0, 13419.0,
                                  13880.0, 14353.0, 14839.0, 15347.0,
                                  15861.0, 16388.0, 16939.0, 17493.0,
                                  18049.0, 18639.0, 19237.0, 19840.0,
                                  20472.0, 21105.0, 21757.0, 22427.0,
                                  23104.0, 23808.0, 24526.0, 25256.0]),

                  'm5': np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
                                  -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
                                  -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
                                  -1.0,-1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 3.0,
                                  4.0, 5.0, 10.1, 18.7, 29.2, 41.7,
                                  54.6,69.0, 93.8, 112.0, 134.2, 155.8,
                                  178.8, 202.3, 227.9, 253.9, 280.0,
                                  307.2,335.2, 368.3, 405.2, 443.9, 484.9,
                                  528.2, 573.0, 619.3, 676.4, 726.6,780.5,
                                  836.0, 883.8, 928.8, 980.4, 1027.0,
                                  1083.4, 1127.5, 1189.6, 1241.1, 1292.0,
                                  1351.0, 1409.0, 1468.0, 1528.0, 1589.0,
                                  1662.0, 1735.0, 1809.0, 1883.0, 1960.0,
                                  2040.0, 2122.0, 2206.0, 2295.0, 2389.0,
                                  2484.0, 2580.0, 2683.0, 2787.0, 2892.0,
                                  3000.0, 3105.0, 3219.0, 3332.0, 3442.0,
                                  3552.0, 3664.0, 3775.0, 3890.0, 4009.0,
                                  4127.0])}

_xraydb = None

def get_xraydb():
    """return instance of the XrayDB

    Returns:
        XrayDB

    Example:
        >>> import xraydb
        >>> xdb = xraydb.get_xraydb()

    """
    global _xraydb
    if _xraydb is None:
        _xraydb = XrayDB()
    return _xraydb

def f0(ion, q):
    """elastic X-ray scattering factor, f0(q), for an ion.

    Args:
       ion (int or str):  atomic number, atomic symbol or ionic symbol of scatterer

       q  (float, ndarray):  Q value(s)  for scattering

    Returns
       scattering factor for each Q value

    Notes:
       1.  from D. Waasmaier and A. Kirfel, Acta Cryst. A51 p416 (1995) and
           International Tables for Crystallography, Vol. C.
       2.  `ion` can be of the form: 26, `Fe`, `Fe2+`.  For a full list of ions
           use `f0_ions()`
       3. elements supported are from Z = 1 to 98 ('H' to 'Cf')
       4. q = sin(theta) / lambda, where theta=incident angle, lambda=X-ray wavelength

    """
    xdb = get_xraydb()
    return xdb.f0(ion, q)

def f0_ions(element=None):
    """list ion names supported in the f0() calculation from
    Waasmaier and Kirfel.

    Args:
      element (None, int, str):  scatterer

    Returns:
      list of strings for matching ion names

    Notes:
        if element is None, all 211 ions are returned.
    """
    xdb = get_xraydb()
    return xdb.f0_ions(element=element)

def chantler_energies(element, emin=0, emax=1.e9):
    """energies at which Chantler data is tabulated for a particular element.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        emin (float):        lower bound of energies (default=0)
        emax (float):        upper bound of energies (default=1.e9)

    Returns:
        ndarray of energies

    Notes:
        energies are in eV
    """
    xdb = get_xraydb()
    return xdb.chantler_energies(element, emin=emin, emax=emax)


def chantler_data(element, energy, column, **kws):
    """data from Chantler tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies
        column (str): data to return, one of 'f1', 'f2', 'mu_photo',
                     'mu_incoh', 'mu_total'

    Returns:
        value or ndarray of values
    """
    xdb = get_xraydb()
    return xdb._from_chantler(element, energy, column=column, **kws)


def f1_chantler(element, energy, **kws):
    """real part of anomalous x-ray scattering factor for an element and
    energy or array of energies.   Data is from the Chantler tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of electrons

    """
    xdb = get_xraydb()
    return xdb.f1_chantler(element, energy, **kws)


def f2_chantler(element, energy):
    """imaginary part of anomalous x-ray scattering factor for an element and
    energy or array of energies.   Data is from the Chantler tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of electrons

    """
    xdb = get_xraydb()
    return xdb.f2_chantler(element, energy)


def mu_chantler(element, energy, incoh=False, photo=False):
    """X-ray mass attenuation coeficient, mu/rho, for an element and
    energy or array of energies.   Data is from the Chantler tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies
        incoh (bool): whether to return only the incoherent contribution [False]
        photo (bool): whether to return only the photo-electric contribution [False]

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of cm^2/gr
        2. The default is to return total attenuation coefficient.
    """
    xdb = get_xraydb()
    return xdb.mu_chantler(element, energy, incoh=incoh, photo=photo)

def mu_elam(element, energy, kind='total'):
    """X-ray mass attenuation coefficient, mu/rho, for an element and
    energy or array of energies.  Data is from the Elam tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies
        kind (str):  type of cross-section to use, one of ('total',
                     'photo', 'coh', 'incoh') ['total']

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of cm^2/gr
        2. The default is to return total attenuation coefficient.

    """
    xdb = get_xraydb()
    return xdb.mu_elam(element, energy, kind=kind)


def coherent_cross_section_elam(element, energy):
    """coherent scaattering cross-section for an element and
    energy or array of energies.  Data is from the Elam tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of cm^2/gr
    """
    xdb = get_xraydb()
    return xdb.coherent_cross_section_elam(element, energy)


def incoherent_cross_section_elam(element, energy):
    """incoherent scaattering cross-section for an element and
    energy or array of energies.  Data is from the Elam tables.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        energy (float or ndarray):   energy or array of energies

    Returns:
        float value or ndarray

    Notes:
        1. Values returned are in units of cm^2/gr
    """
    xdb = get_xraydb()
    return xdb.incoherent_cross_section_elam(element, energy)


def atomic_number(element):
    """z for element name

    Args:
        element (str):  atomic symbol

    Returns:
        atomic number
    """
    xdb = get_xraydb()
    return int(xdb._elem_data(element).Z)


def atomic_symbol(z):
    """atomic symbol for atomic number

    Args:
        z (int):  atomic number

    Returns:
        atomic symbol
    """
    xdb = get_xraydb()
    return xdb._elem_data(z).symbol


def atomic_mass(element):
    """molar mass for an element

    Args:
        element (int, str):  atomic number, atomic symbol for element

    Return:
        atomic mass, in AMU
    """
    xdb = get_xraydb()
    if isinstance(element, int):
        element = atomic_symbol(element)
    return xdb._elem_data(element).mass


def atomic_density(element):
    """density (gr/cm^3) for common for of an element

    Args:
        element (int, str):  atomic number, atomic symbol for element

    Return:
        density in gm/cm^3

    """
    xdb = get_xraydb()
    if isinstance(element, int):
        element = atomic_symbol(element)
    return xdb._elem_data(element).density


def xray_edges(element):
    """get dictionary of x-ray absorption edges:
         energy(in eV),
         fluorescence yield, and
         jump ratio for an element.

    Args:
        element (int, str):  atomic number, atomic symbol for element

    Return:
        dictionary of XrayEdge named tuples.

    Notes:
        1. The dictionary will have keys of edge (iupac symbol)
           and values containing an XrayEdge namedtuple containing
           (energy, fluorescence_yield, edge_jump)
    """
    xdb = get_xraydb()
    return xdb.xray_edges(element)


def xray_edge(element, edge, energy_only=False):
    """get x-ray absorption edge data for an element:
    (energy(in eV), fluorescence yield, jump ratio)

    Args:
        element (int, str):  atomic number, atomic symbol for element
        edge (str): iupac symbol of X-ray edge
        energy_only (bool): whether to return only the energy [False]

    Returns:
         XrayEdge namedtuple containing (energy,
         fluorescence_yield, edge_jump) or float of energy
    """
    xdb = get_xraydb()
    out = xdb.xray_edge(element, edge)
    if energy_only:
        out = out[0]
    return out


def xray_lines(element, initial_level=None, excitation_energy=None):
    """get dictionary of X-ray emission lines of an element

    Args:
        element (int, str):  atomic number, atomic symbol for element
        initial_level (None or str): iupac symbol of initial level
        excitation_energy (None or float): exciation energy

    Returns:
        dict of X-ray lines with keys of siegbahn notation and values of
        XrayLine tuples of (energy, intensity, initial level, final level)


    Notes:
        1. excitation energy will supercede initial_level, as it means
           'all intial levels with below this energy

    Exaample:
        >>> for name, line in xraydb.xray_lines('Mn', 'K').items():
        ...     print(name, line)
        ...
        Ka3 XrayLine(energy=5769.9, intensity=0.000265963, initial_level='K', final_level='L1')
        Ka2 XrayLine(energy=5889.1, intensity=0.293941, initial_level='K', final_level='L2')
        Ka1 XrayLine(energy=5900.3, intensity=0.58134, initial_level='K', final_level='L3')
        Kb3 XrayLine(energy=6491.8, intensity=0.042234, initial_level='K', final_level='M2')
        Kb1 XrayLine(energy=6491.8, intensity=0.0815329, initial_level='K', final_level='M3')
        Kb5 XrayLine(energy=6537.0, intensity=0.000685981, initial_level='K', final_level='M4,5')

    """
    xdb = get_xraydb()
    return xdb.xray_lines(element, initial_level=initial_level,
                          excitation_energy=excitation_energy)


def xray_line(element, line):
    """get data for an  x-ray emission line of an element, given
    the siegbahn notation for the like (Ka1, Lb1, etc).

    Returns:
         energy (in eV), intensity, initial_level, final_level

    Args:
        element (int, str):  atomic number, atomic symbol for element
        line (str):  siegbahn notation for emission line

    Returns:
        an XrayLine namedtuple with (energy, intensity, intial_level, final_level)

    Notes:
       1. if line is not a specifictransition but a generic name like
       'Ka', 'Kb', 'La', 'Lb', 'Lg', without number,  the weighted average for this
       family of lines is returned.

    """
    xdb = get_xraydb()
    lines = xdb.xray_lines(element)

    family = line.lower()
    if family == 'k': family = 'ka'
    if family == 'l': family = 'la'
    if family in ('ka', 'kb', 'la', 'lb', 'lg'):
        scale = 1.e-99
        value = 0.0
        linit, lfinal =  None, None
        for key, val in lines.items():
            if key.lower().startswith(family):
                value += val[0]*val[1]
                scale += val[1]
                if linit is None:
                    linit = val[2]
                if lfinal is None:
                    lfinal = val[3][0]
        return XrayLine(value/scale, scale, linit, lfinal)
    else:
        return lines.get(line.title(), None)


def fluor_yield(element, edge, line, energy):
    """fluorescence yield for an X-ray emission line or family of lines.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        edge (str): iupac symbol of X-ray edge
        line (str): siegbahn notation for emission line
        energy (float): incident X-ray energy

    Returns:
        fluorescence yield, weighted average fluorescence energy, net_probability


    Examples:
        >>> xraydb.fluor_yield('Fe', 'K', 'Ka', 8000)
        0.350985, 6400.752419799043, 0.874576096

        >>> xraydb.fluor_yield('Fe', 'K', 'Ka', 6800)
        0.0, 6400.752419799043, 0.874576096

        >>> xraydb.fluor_yield('Ag', 'L3', 'La', 6000)
        0.052, 2982.129655446868, 0.861899000000000

    See Also:
         `xray_lines` which gives the full set of emission lines ('Ka1', 'Kb3',
         etc) and probabilities for each of these.

    """
    e0, fyield, jump = xray_edge(element, edge)
    trans  = xray_lines(element, initial_level=edge)

    lines = []
    net_ener, net_prob = 0., 0.
    for name, vals in trans.items():
        en, prob = vals[0], vals[1]
        if name.startswith(line):
            lines.append([name, en, prob])

    for name, en, prob in lines:
        if name.startswith(line):
            net_ener += en*prob
            net_prob += prob
    if net_prob <= 0:
        net_prob = 1
    net_ener = net_ener / net_prob
    if energy < e0:
        fyield = 0
    return fyield, net_ener, net_prob


def ck_probability(element, initial, final, total=True):
    """transition probability for an element, initial, and final levels.

    Args:
        element (int, str):  atomic number, atomic symbol for element
        initial (str):  iupac symbol for initial level
        final (str):  iupac symbol for final level
        total (bool): whether to include transitions via possible
                      intermediate levels [True]

    Returns:
        transition probability, or 0 if transition is not allowed.
    """
    xdb = get_xraydb()
    return xdb.ck_probability(element, initial, final, total=total)


def core_width(element, edge=None):
    """returns core hole width for an element and edge

    Args:
        element (int or str): element
        edge (None or str):  edge to consider

    Returns:
        core width or list of core widths

    Notes:
       1. if edge is None, values are return for all edges
       2. Data from Krause and Oliver (1979) and  Keski-Rahkonen and Krause (1974)

    """
    xdb = get_xraydb()
    return xdb.corehole_width(element, edge=edge)

def ionization_potential(gas):
    """return effective ionization potential for a gas or diode semiconductor, as
    appropriate for ionization chambers in the linear regime (not in the
    'proportional counter' regime) or for PIN photodiodes (not in 'avalanche' mode).

    Args:
        gas (string):  name of gas or 'Si' or 'Ge'

    Returns:
        ionization potential in eV

    Notes:
       Data from G. F. Knoll, Radiation Detection and Measurement, Table 5-1, and
       from ICRU Report 31, 1979.  Supported gas names and effective potentials:

           ==================== ================
            gas names            potential (eV)
           -------------------- ----------------
            hydrogen, H           36.5
            helium, He            41.3
            nitrogen, N, N2       34.8
            oxygen, O, O2         30.8
            neon, Ne              35.4
            argon, Ar             26.4
            krypton, Kr           24.4
            xenon, Xe             22.1
            air                   33.8
            methane, CH4          27.3
            carbondioxide, CO2    33.0
            silicon, Si            3.68
            germanium, Ge          2.97
           ==================== ================

       If the gas is not recognized the default value of 32.0 eV will be returned.

    """
    xdb = get_xraydb()
    try:
        return xdb.ionization_potential(gas)
    except:
        return 32.0


def guess_edge(energy, edges=['K', 'L3', 'L2', 'L1', 'M5']):
    """guess an element and edge based on energy (in eV)

    Args:
        energy (float) : approximate edge energy (in eV)
        edges (None or list of strings) : edges to consider

    Returns:
        a tuple of (atomic symbol, edge) for best guess

    Notes:
        by default, the list of edges is ['K', 'L3', 'L2', 'L1', 'M5']

    """
    xdb = get_xraydb()
    ret = []
    min_diff = 1e9

    for edge in edges:
        ename =  edge.lower()
        # if not already in _edge_energies, look it up and save it now

        if ename not in _edge_energies:
            energies = [-1000]*150
            maxz = 0
            for row in xdb.tables['xray_levels'].select().execute().fetchall():
                ir, elem, edgename, en, eyield, xjump = row
                iz = xdb.atomic_number(elem)
                maxz = max(iz, maxz)
                if ename == edgename.lower():
                    energies[iz] = en
            _edge_energies[ename] = np.array(energies[:maxz])

        energies = _edge_energies[ename]
        iz = int(index_nearest(energies, energy))
        diff = energy - energies[iz]
        if diff < 0: # prefer positive errors
            diff = -2.0*diff
        if iz < 10 or iz > 92: # penalize extreme elements
            diff = 2.0*diff
        if edge == 'K': # prefer K edge
            diff = 0.25*diff
        elif edge in ('L1', 'M5'): # penalize L1 and M5 edges
            diff = 2.0*diff
        if diff < min_diff:
            min_diff = diff
        ret.append((edge, iz, diff))

    for edge, iz, diff in ret:
        if abs(diff - min_diff) < 2:
            return (atomic_symbol(iz), edge)
    return (None, None)


class Scatterer:
    """Scattering Element

    lamb=PLANCK_HC /(eV0/1000.)*1e-11    # in cm, 1e-8cm = 1 Angstrom
    Xsection=2* R_ELECTRON_CM *lamb*f2/BARN    # in Barns/atom
    """
    def __init__(self, symbol, energy=10000):
        # atomic symbol and incident x-ray energy (eV)
        self.symbol = symbol
        self.number = atomic_number(symbol)
        self.mass   = atomic_mass(symbol)
        self.f1     = self.number + f1_chantler(symbol, energy)
        self.f2     = f2_chantler(symbol, energy)
        self.mu_photo = chantler_data(symbol, energy, 'mu_photo')
        self.mu_total = chantler_data(symbol, energy, 'mu_total')

def xray_delta_beta(material, density, energy):
    """anomalous components of the index of refraction for a material,
    using the tabulated scattering components from Chantler.

    Args:
       material:   chemical formula  ('Fe2O3', 'CaMg(CO3)2', 'La1.9Sr0.1CuO4')
       density:    material density in g/cm^3
       energy:     x-ray energy in eV

    Returns:
      (delta, beta, atlen)

    where
      delta :  real part of index of refraction
      beta  :  imag part of index of refraction
      atlen :  attenuation length in cm

    These are the anomalous scattering components of the index of refraction:

    n = 1 - delta - i*beta = 1 - lambda**2 * r0/(2*pi) Sum_j (n_j * fj)

    Adapted from code by Yong Choi

    """

    lamb_cm = 1.e-8 * PLANCK_HC / energy # lambda in cm
    elements = []

    for symbol, number in chemparse(material).items():
        elements.append((number, Scatterer(symbol, energy)))

    total_mass, delta, beta_photo, beta_total = 0, 0, 0, 0
    for (number, scat) in elements:
        weight      = density*number*AVOGADRO
        delta      += weight * scat.f1
        beta_photo += weight * scat.f2
        beta_total += weight * scat.f2*(scat.mu_total/scat.mu_photo)
        total_mass += number * scat.mass

    scale = lamb_cm * lamb_cm * R_ELECTRON_CM / (2*np.pi*total_mass)
    delta  *= scale
    beta_photo *= scale
    beta_total *= scale
    if isinstance(beta_total, np.ndarray):
        beta_total[np.where(beta_total<1.e-99)] = 1.e-99
    else:
        beta_total = max(beta_total, 1.e-19)
    return delta, beta_photo, lamb_cm/(4*np.pi*beta_total)

def mirror_reflectivity(formula, theta, energy, density=None,
                        roughness=0.0, polarization='s'):
    """mirror reflectivity for a thick, singl-layer mirror.

    Args:
       formula (string):           material name or formula ('Si', 'Rh', 'silicon')
       theta (float or nd-array):  mirror angle in radians
       energy (float or nd-array): X-ray energy in eV
       density (float or None):    material density in g/cm^3
       roughness (float):          mirror roughness in Angstroms
       polarization ('s' or 'p'):  mirror orientation relative to X-ray polarization

    Returns:
       mirror reflectivity values

    Notes:
       1. only one of theta or energy can be an nd-array
       2. density can be `None` for known materials
       3. polarization of 's' puts the X-ray polarization along the mirror
          surface, 'p' puts it normal to the mirror surface. For
          horizontally polarized X-ray beams from storage rings, 's' will
          usually mean 'vertically deflecting' and 'p' will usually mean
          'horizontally deflecting'.
    """
    from .materials import get_material
    if density is None:
        formula, density = get_material(formula)

    delta, beta, _ = xray_delta_beta(formula, density, energy)
    n = 1 - delta - 1j*beta

    # kiz is k in air/vacuum,  with n = 1.
    # ktz is k in mirror material, with n < 1.
    qf  = 2*np.pi * energy/PLANCK_HC
    kiz = qf * np.sin(theta)
    ktz = qf * np.sqrt(n**2 - np.cos(theta)**2)

    # polarization correction will be tiny for small angles
    if polarization == 'p':
        ktz = ktz / n

    r_amp = (kiz - ktz)/(kiz + ktz)
    if roughness > 1.e-12:
        r_amp = r_amp * np.exp(-2*(roughness**2*kiz*ktz))
    return (r_amp*r_amp.conjugate()).real


def ionchamber_fluxes(gas='nitrogen', volts=1.0, length=100.0,
                      energy=10000.0, sensitivity=1.e-6,
                      sensitivity_units='A/V', with_compton=1):

    """return ion chamber and PIN diode fluxes for a gas, mixture of gases, or
    semiconductor material, ion chamber length (or diode thickness), X-ray energy,
    recorded voltage and current amplifier sensitivity.  See note for details.

    Args:
        gas (string or dict):  name or formula of fill gas (see note 1) ['nitrogen']
        volts (float):  measured voltage output of current amplifier  [1.0]
        length (float): active length of ion chamber in cm [100]
        energy (float): X-ray energy in eV [10000]
        sensitivity (float): current amplifier sensitivity [1.e-6]
        sensitivity_units (string): units of current amplifier sensitivity
                                    (see note 2 for options) ['A/V']
        with_compton (int): switch to control the contribution of Compton
                            scattering (see note 3) [1]


    Returns:
        named tuple IonchamberFluxes with fields

            `incident`    flux of beam incident on ion chamber in Hz

            `transmitted` flux of beam output of ion chamber in Hz

            `photo`       flux absorbed by photo-electric effect in Hz

            `incoherent`  flux attenuated by incoherent scattering in Hz


    Examples:

        >>> from xraydb import ionchamber_fluxes
        >>> fl = ionchamber_fluxes(gas='helium', volts=1.25, length=20.0,
                                   energy=10000.0, sensitivity=1.e-9)

        >>> print(f"Fluxes: In={fl.incident:g}, Out={fl.transmitted:g}, Transmitted={100*fl.transmitted/fl.incident:.2f}%")
        Fluxes: In=2.79631e+11, Out=2.79383e+11, Transmitted=99.91%

        >>> fl = ionchamber_fluxes(gas='nitrogen', volts=1.25, length=20.0,
                                   energy=10000.0, sensitivity=1.e-6)

        >>> print(f"Fluxes: In={fl.incident:g}, Out={fl.transmitted:g}, Transmitted={100*fl.transmitted/fl.incident:.2f}%")
        Fluxes: In=3.20045e+11, Out=2.90464e+11, Transmitted=90.76%

        >>> fl = ionchamber_fluxes(gas={'nitrogen':0.5, 'helium': 0.5},
                                  volts=1.25, length=20.0, energy=10000.0,
                                  sensitivity=1.e-6)

        >>> print(f"Fluxes: In={fl.incident:g}, Out={fl.transmitted:g}, Transmitted={100*fl.transmitted/fl.incident:.2f}%")
        Fluxes: In=6.83845e+11, Out=6.51188e+11, Transmitted=95.22%

    Notes:
       1. The gas value can either be a string for the name of chemical
          formula for the gas or diode material, or dictionary with keys
          that are gas names or formulas and values that are the relative
          fraction for mixes gases.  For diode materials, mixtures are not
          supported.

          The gas formula is used both the contributions for mu and to get
          get the weighted effective ionization potential for the material.

          The effective ionization potentials are known for a handful of
          gases and diodes (see `ionization_potential` function), and range
          between 20 and 45 eV for gases, and around 3 eV for semiconductors.
          For unknown gases the value of 32.0 eV will be used.

       2. The `sensitivity` and `sensitivity_units` arguments have some overlap
          to specify the sensitivity or gain of the current amplifier.
          Generally, the units in `A/V`, but you can add a common SI prefix of
          'p', 'pico', 'n', 'nano', (unicode 'u03bc'), 'u', 'micro', 'm', 'milli'
          so that, `ionchamber_fluxes(..., sensitivity=1.e-6)` and
          `ionchamber_fluxes(..., sensitivity=1, sensitivity_units='uA/V')`
          will both give a sensitivity of 1 microAmp / Volt .

       3. The effect of Compton scattering on the ion chamber current can be approximated
          in 3 ways with different values for `with_compton` (default=1):

             >0:  use Compton-shifted electron energy (best approximation)

             =0:  completely ignore the effect of Compton scattering on current

             <0:  use incident energy as Compton energy.

          Using -1 should reproduce the calculation from Hephaestus reasonably
          well, but using 1 is highly recommended.

       4. Ion Chamber Current calculation: the total attenuation in an ion
          chamber includes photo, incoherent (Compton), and coherent (Rayleigh)
          scattering contributions:

            Flux_transmitted = Flux_in * exp(-t*mu_total)

          The current in an Ion Chamber has a contribution from both the
          photo-electric and incoherent (Compton) cross-section, but not
          the coherent scattering. For an incident flux (in Hz) of Flux_in,
          the fluxes of X-rays attenuated by the different processes are:

             Flux_photo = Flux_in * [1 - exp(-t*mu_photo)]

             Flux_incoh = Flux_in * [1 - exp(-t*mu_incoh)]

             Flux_coh   = Flux_in * [1 - exp(-t*mu_coh)]

             Flux_transmitted = Flux_in - Flux_total

          For Flux_photo, all of the X-ray energy is converted to current:

             IC_current_photo = Flux_photo * Energy * q_e / ion_pot

          where q_e is the electron charge (1.602e-19 C), and ion_pot is the
          effective ionization potential (see `ionization_potential`).

          For Flux_incoh, the energy transferred to the electron is converted
          to electron and ion current as:

             IC_current_incoh = Flux_incoh * Energy_Compton * q_e / ion_pot

          where:

             Energy_Compton = Energy/(1+m_e*c^2/Energy)

          is the approximate energy of the Compton-scattered electron.  This
          estimate of the energy is a simplification: the transferred energy will
          be angle-dependent, ranging from 0 to 2*Energy_Compton, with a
          distribution that depends on Energy and polarization.

    """
    from .materials import material_mu

    fin = fout = fphoto = 0.0

    units = sensitivity_units.replace('Volts', 'V').replace('Volt', 'V')
    units = units.replace('Amperes', 'A').replace('Ampere', 'A')
    units = units.replace('Amps', 'A').replace('Amp', 'A')
    units = units.replace('A/V', '')
    sensitivity *= SI_PREFIXES.get(units, 1)

    if isinstance(gas, str):
        gas = {gas: 1.0}

    gas_total = 0.0
    gas_comps = []
    for gname, frac in gas.items():
        ionpot = ionization_potential(gname)
        if gname == 'N2': gname = 'nitrogen'
        if gname == 'O2': gname = 'oxygen'
        gas_total += frac
        gas_comps.append((gname, frac, ionpot))

    # energy of Compton-scattered electron: median energy, approximate
    # can turn off Compton energy with with_compton switch:
    energy_compton = 0      # no Compton contribution
    if with_compton > 0:    # median Compton energy
        energy_compton = energy/(1 + E_MASS/energy)
    elif with_compton < 0:  # use incident X-ray energy: hephaestus
        energy_compton = energy

    # use weighted sums for mu values and ionization potential
    mu_photo, mu_incoh, mu_total, ion_pot =  0.0, 0.0, 0.0, 0.0
    for gas_name, gas_frac, gas_ion_pot in gas_comps:
        gasmu_photo = material_mu(gas_name, energy=energy, kind='photo')
        gasmu_total = material_mu(gas_name, energy=energy, kind='total')
        gasmu_incoh = material_mu(gas_name, energy=energy, kind='incoh')

        mu_photo += gasmu_photo * gas_frac/gas_total
        mu_total += gasmu_total * gas_frac/gas_total
        mu_incoh += gasmu_incoh * gas_frac/gas_total
        ion_pot  += gas_ion_pot * gas_frac/gas_total

    absorbed_energy = (energy*(1-np.exp(-length* mu_photo)) +
                       energy_compton*(1-np.exp(-length* mu_incoh)))
    flux_in    = volts*sensitivity*ion_pot/(QCHARGE*absorbed_energy)
    flux_photo = flux_in * (1-np.exp(-length* mu_photo))
    flux_incoh = flux_in * (1-np.exp(-length* mu_incoh))
    flux_out   = flux_in * np.exp(-length*mu_total)

    return fluxes(incident=flux_in, transmitted=flux_out,
                  photo=flux_photo, incoherent=flux_incoh)


def darwin_width(energy, crystal='Si', hkl=(1, 1, 1), a=None,
                 polarization='s', ignore_f2=False, ignore_f1=False, m=1):

    """darwin width for a crystal reflection and energy

    Args:
      energy (float):    X-ray energy in eV
      crystal (string):  name of crystal (one of 'Si', 'Ge', or 'C') ['Si']
      hkl (tuple):       h, k, l for reflection  [(1, 1, 1)]
      a (float or None): lattice constant [None - use built-in value]
      polarization ('s','p', 'u'): mono orientation relative to X-ray polarization ['s']
      ignore_f1 (bool):  ignore contribution from f1 - dispersion (False)
      ignore_f2 (bool):  ignore contribution from f2 - absorption (False)
      m (int):           order of reflection    [1]

    Returns:

      A named tuple 'DarwinWidth' with the following fields

        `theta`:        float, nominal Bragg angle, in rad,

        `theta_offset`: float, angular offset from Bragg angle, in rad,

        `theta_width`:  float, estimated angular Darwin width, in rad,

        `theta_fwhm`:   float, estimated FWHM of angular intensity, in rad,

        `energy_width`: float, estimated angular Darwin width, in rad,

        `energy_fwhm`:  float, estimated FWHM of energy intensity, in eV,

        `zeta`:         nd-array of Zeta parameter (delta_Lambda / Lambda),

        `dtheta`:       nd-array of angles away from Bragg angle, theta in rad,

        `denergy`:      nd-array of energies away from Bragg energy, in eV,

        `intensity`:    nd-array of reflected intensity

    Notes:

     1. This follows the calculation from section 6.4 of
        Elements of Modern X-ray Physics, 2nd Edition
        J Als-Nielsen, and D. McMorrow.

     2. Only diamond structures (Si, Ge, diamond) are currently supported.
        Default values of lattice constant `a` are in Angstroms: 5.4309 for Si,
        5.6578, for 'Ge', and 3.567 for 'C'.

     3. The `theta_width` and `energy_width` values will closely match the
        width of the intensity profile that would = 1 when ignoring the
        effect of absorption.  These are the values commonly reported as
        'Darwin Width'.  The value reported for `theta_fwhm' and
        `energy_fwhm` are larger than this by sqrt(9/8) ~= 1.06.

     4. Polarization can be 's', 'p', 'u',  or None. 's' means vertically
        deflecting crystal and a horizontally-polarized source, as for most
        synchrotron beamlines. 'p' is for a horizontally-deflecting crystal.
        'u' or None is for unpolarized light, as for most fluorescence/emission.

    Examples:
        >>> dw = darwin_width(10000, crystal='Si', hkl=(1, 1, 1))
        >>> dw.theta_width, dw.energy_width
        (2.8593683930207114e-05, 1.4177346002236872)

    """
    lattice_constants = {'Si': 5.4309, 'Ge': 5.6578, 'C': 3.567}

    h_, k_, l_ = hkl
    hklsum = (h_ + k_ + l_)
    if hklsum % 4 == 0 and (h_ % 2 == 0 and k_ % 2 == 0 and l_ % 2 == 0):
        eqr = 8
    elif (h_ % 2 == 1 and k_ % 2 == 1 and l_ % 2 == 1): # all odd
        eqr =4*np.sqrt(2)
    else:
        raise ValueError("hkl must sum to 4 or be all odd")

    if a is None:
        a = lattice_constants[crystal.title()]
    dspace = a / np.sqrt(h_*h_ + k_*k_ + l_*l_)
    lambd  = PLANCK_HC / energy
    if lambd > 2*dspace:
        return DarwinWidth(theta=np.nan, theta_offset=np.nan,
                           theta_width=np.nan, theta_fwhm=np.nan,
                           energy_width=np.nan, energy_fwhm=np.nan,
                           zeta=[], dtheta=[], denergy=[], intensity=[])

    theta  = np.arcsin(lambd/(2*dspace))
    q  = 0.5 / dspace
    f1 = f2 = 0
    if not ignore_f1:
        f1 = f1_chantler(crystal, energy)
    if not ignore_f2:
        f2 = f2_chantler(crystal, energy)

    gscale = 2 * (dspace)**2 * R0 / (m*a**3)

    if polarization is None or polarization.startswith('u'): # unpolarized
        gscale *= (1 + abs(np.cos(2*theta)))/2.0
    elif polarization.startswith('p'):
        gscale *= abs(np.cos(2*theta))

    g0 = gscale * 8   * (f0(crystal, 0)[0] + f1 - 1j*f2)
    g  = gscale * eqr * (f0(crystal, q)[0] + f1 - 1j*f2)

    total = abs(2*g/(m*np.pi))
    fwhm  = total * 3/(2*np.sqrt(2))  # where do A-N&M get this factor?

    zeta_offset = g0.real/np.pi
    theta_offset = np.tan(theta)*zeta_offset

    # as a check, the following formula from L Berman (and X0h doc)
    # will give identical results as theta_width. [sin(2x)= 2sin(x)*cos(x)]
    # dw_lb = 2*R0*lambd**2 * eqr*abs(f0(crystal, q)[0] + f1 - 1j*f2)/(m*np.pi*a**3* np.sin(2*theta))

    #  hueristic zeta range and step sizes for crystals:
    sz =  zeta_offset

    zeta = np.concatenate((np.arange(-1.5*sz,    0, 0.05*total),
                           np.arange(0,       2*sz, 0.01*total),
                           np.arange(2*sz,  3.5*sz, 0.05*total)))
    xc = (m*np.pi*zeta - g0)/g
    _p = np.where(xc.real > 1)[0]
    _n = np.where(xc.real < -1)[0]

    r     = (xc - 1j * np.sqrt(1 - xc**2))
    r[_p] = (xc - np.sqrt(xc**2 -1))[_p]
    r[_n] = (xc + np.sqrt(xc**2 -1))[_n]

    return DarwinWidth(theta=theta,
                       theta_offset=theta_offset,
                       theta_width=total*np.tan(theta),
                       theta_fwhm=fwhm*np.tan(theta),
                       energy_width=total*energy,
                       energy_fwhm=fwhm*energy,
                       zeta=zeta,
                       dtheta=zeta*np.tan(theta),
                       denergy=-zeta*energy,
                       intensity=abs(r*r.conjugate()))


def transmission_sample(sample, energy, absorp_total=2.6, area=1, 
                        density=None, frac_type='mass'):
    """Analyze transmission mode sample. Sample can be specified as a chemical
    formula (str or dict) or as mass fractions (dict). One mass fraction can 
    have value -1 to indicate the unspecified portion of the mass fractions 
    (i.e. so made to equal one).

    Absorbance steps for each element are calculated. This is done by performing
    a polynomial fit to the pre-edge absorption (from -200 to -50 eV of specified energy) 
    and extrapolating that to the post-edge, then comparing against the actual 
    computed post-edge absorption.

    Args:
        sample (str or dict): elements/compounds and their mass fractions. one 
                           entry can have value -1 to indicate unspecified portion
        energy (float): X-ray energy (eV) at which transmission will be analyzed
        absorp_total (float): total absorption (mu_t*d) of the sample at the
                              specified energy
        area (float)(optional): area (cm^2) of the sample. Default is 1 cm^2.
        density (float)(optional): density (g/cm^3) of the sample
        frac_type (str)(optional): can be `mass` or `molar`, if sample is dict, 
                                   this keyword specifies whether the indicated 
                                   fractions are mass fractions or 
                                   molar fractions (i.e. chemical formula)

    Returns:
        dictionary with fields

            `energy(eV)`        incident energy

            `absorp_total`      total absorption

            `mass_fractions`    mass fractions of elements

            `absorbance_steps`  absorbance steps of each element in the sample

            `area (cm^2)`       area, if specified

            `mass_total(mg)`    total mass of sample (if area specified)

            `mass_components(mg)`   mass of each element (if area specified)

            `density(g/cc)`     density, if specified

            `thickness(mm)`     thickness of sample (if area AND density specified)

            `absorption_length(um)` abs. length of sample (if area AND density specified)

    Examples:

        5% Fe in Silica
        >>> transmission_sample(
                sample={'Fe': 0.05, 'SiO2': -1},
                energy=xraydb.xray_edge('Fe', 'K').energy + 50,
                area=1.33,
                frac_type='mass',
            )
        
        Output:
            TransmissionSample(
                energy_eV=7162.0,
                absorp_total=2.6,
                mass_fractions={
                    'Fe': 0.05,
                    'Si': 0.4440648769202603,
                    'O': 0.5059351230797396},
                absorbance_steps={
                    'Fe': 0.6692395733146204,
                    'Si': 1.297403071978392e-06,
                    'O': 3.386553723091669e-07},
                area_cm2=1.33,
                mass_total_mg=51.05953690489308,
                mass_components_mg={
                    'Fe': 2.552976845244654,
                    'Si': 22.673746971276834,
                    'O': 25.832813088371587},
                density=None,
                thickness_mm=None,
                absorption_length_um=None
            )

        >>> transmission_sample(
                sample='Fe2O3',
                energy=xraydb.xray_edge('Fe', 'K').energy + 50,
                area=1.33
            )

        Output:
            TransmissionSample(
                energy_eV=7162.0,
                absorp_total=2.6,
                mass_fractions={
                    'Fe': 0.6994307614270416,
                    'O': 0.3005692385729583},
                absorbance_steps={
                    'Fe': 2.2227981005407176,
                    'O': 4.7769571901536886e-08},
                area_cm2=1.33,
                mass_total_mg=12.123291571370844,
                mass_components_mg={
                    'Fe': 8.479403054765946,
                    'O': 3.643888516604898},
                density=None,
                thickness_mm=None,
                absorption_length_um=None
            )

        >>> transmission_sample(
                sample={'Fe': 2, 'O': 3},
                energy=xraydb.xray_edge('Fe', 'K').energy + 50,
                area=1.33,
                frac_type='molar'
            )
        
        Output same as previous example.
    """
    if type(sample) is str:
        sample = formula_to_mass_fracs(sample)
    if type(sample) is dict:
        if frac_type == 'mass':
            sample = _validate_mass_fracs(sample)
        elif frac_type == 'molar':
            sample = formula_to_mass_fracs(sample)
        else:
            raise RuntimeError('`frac_type` must be `mass` or `molar`')
    mu_tot = sum([mu_elam(k, energy) * v for k, v in sample.items()])
    rho_d = absorp_total / mu_tot

    absorbance_steps = {}
    pre_edge = np.linspace(energy - 200, energy - 50, 100)
    for el in sample.keys():
        coeffs = np.polyfit(pre_edge, mu_elam(el, pre_edge), 3)
        extrapolated = sum([c * energy ** (len(coeffs) - 1 - i) \
                            for i, c in enumerate(coeffs)])
        post_edge = mu_elam(el, energy)
        absorbance_steps[el] = (post_edge - extrapolated) * sample[el] * rho_d

    mass_total = None
    mass_components_mg = None
    thickness_mm = None
    absorption_length_um = None

    if area:
        mass_total = rho_d * area * 1000 # mg
        mass_components_mg = {k: v * mass_total for k, v in sample.items()}
        if density:
            thickness_mm = mass_total / (area * 100) / density
            absorption_length_um = 1 / density / mu_tot * 1e4

    return TransmissionSample(
                            energy_eV=energy,
                            absorp_total=absorp_total,
                            mass_fractions=sample,
                            absorbance_steps=absorbance_steps,
                            area_cm2=area,
                            mass_total_mg=mass_total,
                            mass_components_mg=mass_components_mg,
                            density=density,
                            thickness_mm=thickness_mm,
                            absorption_length_um=absorption_length_um
                            )


def formula_to_mass_fracs(formula):
    """Calculate mass fractions of elements from a given molecular formula.

    Args:
        formula (str or dict): chemical formula
    
    Returns:
        dict with fields of each element and values of their mass fractions

    Example:
        >>> formula_to_mass_fracs('Fe2O3')
        {'Fe': 0.6994307614270416, 'O': 0.3005692385729583}

        >>> formula_to_mass_fracs({'Fe': 2, 'O': 3})
        {'Fe': 0.6994307614270416, 'O': 0.3005692385729583}
    """
    if type(formula) is str:
        simplified_formula = chemparse(formula)
    elif type(formula) is dict:
        simplified_formula = {}
        for k, v in formula.items():
            elements = chemparse(k) # handle case of compound formula, e.g. {'Mn':1, 'SiO2':3}
            elements = {el : c * v for el, c in elements.items()}
            for el, c in elements.items():
                if el in simplified_formula:
                    simplified_formula[el] += c
                else:
                    simplified_formula[el] = c
    else:
        raise ValueError('`formula` must have type `str` or `dict`')
    mol_weight = sum([v * atomic_mass(k) for k, v in simplified_formula.items()])
    mass_fracs = {k: v * atomic_mass(k) / mol_weight for k, v in simplified_formula.items()}
    return mass_fracs


def mass_fracs_to_formula(mass_fracs):
    """Calculate molecular formula from a given  mass fractions of elements.

    Args:
        mass_fracs (dict): mass fractions of elements
    
    Returns:
        dict with fields of each element and values of their coefficients

    Example:
        >>> mass_fracs_to_formula({'Fe': 0.7, 'SiO2': -1})
        {
            'Fe': 0.012534694242994, 
            'Si': 0.004993092888171364, 
            'O': 0.009986185776342726
        }
    """
    mass_fracs = _validate_mass_fracs(mass_fracs)
    masses = {}
    for el, _ in mass_fracs.items():
        parsed = chemparse(el)
        masses[el] = sum([atomic_mass(el) * c for el, c in parsed.items()])

    coeffs = {k: mass_fracs[k] / masses[k] for k in mass_fracs.keys()}

    return coeffs


def _validate_mass_fracs(mass_fracs):
    """Validate mass fractions. Either verify they sum to one, or calculate
    the remaining portion of a compound/element with value specified as -1.

    Additionally, compounds specified in mass_fracs will be separated to the
    individual elements.
    """
    if any([v == -1 for v in mass_fracs.values()]):
        unknown = [k for k, v in mass_fracs.items() if v == -1]
        assert len(unknown) == 1, 'Multiple unknown weight percentages'
        mass_fracs[unknown[0]] = 1 - sum({k:v for k, v in mass_fracs.items() if k != unknown[0]}.values())
    else:
        compare = abs(sum([v for v in mass_fracs.values()]) - 1) < 1e-4
        if not compare:
            raise RuntimeError("Mass fractions do not add up to one.")
    
    simplified_mass_fracs = {}
    for comp, frac in mass_fracs.items():
        parsed = chemparse(comp)
        parsed_sum = sum([atomic_mass(el) * c for el, c in parsed.items()])
        for el, c in parsed.items():
            simplified_mass_fracs[el] = atomic_mass(el) * c / parsed_sum * frac
    return simplified_mass_fracs



#!/usr/bin/env python
# XrayDB example script python/examples/mu_components_C.py
#
# plot components of X-ray mass attenuation for C 
#
import numpy as np
import matplotlib.pyplot as plt
from xraydb import mu_elam 

energy = np.arange(500, 120000, 10)  # energy in eV
      
elem = 'C'
mu_total = mu_elam(elem, energy, kind='total')
mu_photo = mu_elam(elem, energy, kind='photo')
mu_incoh = mu_elam(elem, energy, kind='incoh')
mu_coher = mu_elam(elem, energy, kind='coh')

plt.title('X-ray mass attenuation for %s' % elem)
plt.plot(energy, mu_total, linewidth=2, label='Total')
plt.plot(energy, mu_photo, linewidth=2, label='Photo-electric')
plt.plot(energy, mu_incoh, linewidth=2, label='Incoherent')
plt.plot(energy, mu_coher, linewidth=2, label='Coherent')

plt.xlabel('Energy (eV)')
plt.ylabel(r'$\mu/\rho \rm\, (cm^2/gr)$')
plt.legend()
plt.yscale('log')
plt.show()



#!/usr/bin/env python
# XrayDB example script python/examples/mu_elements.py
#
# plot X-ray mass attenuation for selected elements
#
import numpy as np
import matplotlib.pyplot as plt
import wxmplot.interactive as wi
from xraydb import mu_elam , atomic_symbol

energy = np.arange(500, 120000, 10)  # energy in eV
      
for elem in ('C', 'Cu', 'Au'):
    mu = mu_elam(elem, energy)
    plt.plot(energy, mu, label=elem, linewidth=2)
 
plt.title('X-ray mass attenuation')
plt.xlabel('Energy (eV)')
plt.ylabel(r'$\mu/\rho \rm\, (cm^2/gr)$')
plt.legend()
plt.yscale('log')
plt.xscale('log')
plt.show()



import numpy as np
from xraydb import mirror_reflectivity
import matplotlib.pyplot as plt

energy = np.linspace(1000, 51000, 501)

r_si = mirror_reflectivity('Si', 0.002, energy)
r_ni = mirror_reflectivity('Ni', 0.002, energy)
r_rh = mirror_reflectivity('Rh', 0.002, energy)
r_pt = mirror_reflectivity('Pt', 0.002, energy)

plt.plot(energy, r_si, label='Si')
plt.plot(energy, r_ni, label='Ni')
plt.plot(energy, r_rh, label='Rh')
plt.plot(energy, r_pt, label='Pt')

plt.title('X-ray reflectivity at $\\theta=2 \mathrm{mrad}$')
plt.xlabel('Energy (eV)')
plt.ylabel('Reflectivity')
plt.legend()
plt.show()



"""Python interface to XrayDB:

   X-ray Reference Data for the Elements

   last update: 30-August-2019
   License: Public Domain
   Author:  Matthew Newville <newville@cars.uchicago.edu>
            Center for Advanced Radiation Sources,
            The University of Chicago
"""


from .xraydb import XrayDB

from .chemparser import chemparse, validate_formula

from .materials import (material_mu, material_mu_components, get_materials,
                        get_material, find_material, add_material)

from .xray import (atomic_number, atomic_symbol, atomic_mass,
                   atomic_density, xray_edges, xray_edge, xray_lines,
                   xray_line, fluor_yield, ck_probability, core_width, f0,
                   f0_ions, chantler_energies, f1_chantler, f2_chantler,
                   mu_chantler, mu_elam, coherent_cross_section_elam,
                   incoherent_cross_section_elam, guess_edge,
                   xray_delta_beta, get_xraydb, darwin_width,
                   mirror_reflectivity, ionchamber_fluxes,
                   ionization_potential, transmission_sample)

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
